/******************************************************************************    
 * FILE PURPOSE:     - LED driver module Source                                       
 ******************************************************************************    
 * FILE NAME:     led_driver.c                                                        
 *                                                                                 
 * DESCRIPTION:   Linux LED character driver implementation                                                  
 *                                                                                 
 *                                                                                 
 * (C) Copyright 2004, Texas Instruments, Inc                                      
 *******************************************************************************/  
                                                                                   
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/delay.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/wait.h>
#include <linux/proc_fs.h>

#include <asm/avalanche/generic/pformCfg.h>
#include <asm/avalanche/generic/pal.h>
#include "led_platform.h"
#include "led_drv_user.h"
#include "led_hal_drv.h"
#define TI_LED_VERSION          "1.0"

struct led_device {
    wait_queue_head_t input_status_queue;
    int status_change;
    char buffer[128];
    spinlock_t lock;    
};

/* Global variable */
static int gpio_off_state = AVALANCHE_GPIO_OFF_MAP;
static struct led_device *led_dev = NULL;

/* IOCTL function */
static int led_drv_ioctl( struct inode * inode, struct file * file, 
  unsigned int cmd, unsigned long arg )
{
  int ret = 0;
    
  switch ( cmd )
  {
    case LED_CMD_GET_HANDLE:
    {
      struct led_gpio_object led_mod;

      if (copy_from_user((char *)&led_mod, (char *)arg, sizeof(struct led_gpio_object)))
      {
        ret =  -EFAULT;
        break;
      }

      ret = (int)led_hal_user_register(&led_mod);

      if (ret < 0)
        break;
      else
      {  
        /* Save LED handle */
        led_mod.led_handle = ret;  
        ret = 0;
      }  

      if (copy_to_user((char *)arg, (char *)&led_mod, sizeof(struct led_gpio_object)))
      {
        ret =  -EFAULT;
        break;
      }
    }
    break;

    case LED_CMD_ACTION:
    {
      struct led_action led_action;
      
      if (copy_from_user((char *)&led_action, (char *)arg, sizeof(struct led_action)))
      {
        ret =  -EFAULT;
        break;
      }
      else
        ret = led_hal_action(&led_action);
    }
    break;

    case LED_CMD_RELEASE_HANDLE:
      ret = led_hal_deregister((int)arg);
      break;
        
    case LED_CMD_GET_INPUT_HANDLE:
    {
      struct input_gpio_object input_mod;

      if (copy_from_user((char *)&input_mod, (char *)arg, sizeof(struct input_gpio_object)))
      {
        ret =  -EFAULT;
        break;
      }

      ret = (int)led_hal_input_register(&input_mod);
      if (ret < 0)
        break;
      else
      {  
        /* Save input GPIO handle */
        input_mod.input_handle = ret;  
        ret = 0;
      }  

      if (copy_to_user((char *)arg, (char *)&input_mod, sizeof(struct input_gpio_object)))
      {
        ret =  -EFAULT;
        break;
      }
    }
    break;

    case LED_CMD_RELEASE_INPUT_HANDLE:
      ret = led_hal_input_deregister((int)arg);
      break;
    
    case LED_CMD_CONTROL:
      ret = led_hal_control((int)arg);
      break;

    default:
      ret = -EINVAL;
  }
  return ret;
}

static ssize_t led_drv_read(struct file *file, char *buf, size_t count, loff_t *f_pos)
{
  struct led_device *priv_led_dev = file->private_data;
  struct gpio_input_state input_gpio;
  int len = 0, i;
  int flags;
    
  if (f_pos != &file->f_pos) 
    return -ESPIPE;
      
  while (!priv_led_dev->status_change) {
    if (wait_event_interruptible(priv_led_dev->input_status_queue, (priv_led_dev->status_change)))
      return -ERESTARTSYS;
  }

  spin_lock_irqsave(&priv_led_dev->lock, flags);
  for (i = 0; i < LED_MAX_GPIO_INPUTS; i++) {
    if ((led_hal_get_input_status(i, &input_gpio) == 0) && (input_gpio.state_change))
    {
      len += sprintf(priv_led_dev->buffer + len, "GPIO %d: %d\n", 
                input_gpio.gpio_num, input_gpio.gpio_state );
      input_gpio.state_change = 0;
    }
  }    
  priv_led_dev->status_change = 0;
  spin_unlock_irqrestore(&priv_led_dev->lock, flags);
    
  if (len > count)
    len = count;
      
  if (copy_to_user(buf, priv_led_dev->buffer, len))
    return -EFAULT;    
  return len;                                                                                
}

static unsigned int led_drv_poll(struct file *file, poll_table *wait)
{
  struct led_device *priv_led_dev = file->private_data;
    
  poll_wait(file, &priv_led_dev->input_status_queue, wait);
  return (priv_led_dev->status_change ? POLLIN|POLLRDNORM : 0);
}

static int led_drv_open( struct inode * inode, struct file * file )
{
  file->private_data = led_dev;
  file->f_pos = led_dev->status_change;
  return 0;
}

static int led_drv_close( struct inode * inode, struct file * file )
{
  struct led_device *priv_led_dev = file->private_data;
  int flags;
    
  spin_lock_irqsave(&priv_led_dev->lock, flags);
  priv_led_dev->status_change = 0;
  spin_unlock_irqrestore(&priv_led_dev->lock, flags);

  return 0;
}

struct file_operations led_drv_fops = {
  read:      led_drv_read,
  poll:      led_drv_poll,
  ioctl:    led_drv_ioctl,
  open:    led_drv_open,
  release:  led_drv_close
};

void led_drv_notify_state_change(void *drv_data)
{
  struct led_device *priv_led_dev = (struct led_device *)drv_data;
  int flags;
    
  spin_lock_irqsave(&priv_led_dev->lock, flags);
  priv_led_dev->status_change = 1;
  spin_unlock_irqrestore(&priv_led_dev->lock, flags);    
  wake_up_interruptible(&priv_led_dev->input_status_queue);
}

/* Proc function to display driver version */
static int led_drv_ver_info(char *buf, char **start, off_t offset, int count, 
  int *eof, void *data)
{                                                                                              
	int len=0;                                                                                             
	
  len += sprintf(buf +len,"\nTI Linux LED Driver Version %s\n",TI_LED_VERSION);
	return len;                                                                                
}                

/* proc interface /proc/avalanche/led  */
int led_drv_cfg_info(char* buf, char **start, off_t offset, int count,
  int *eof, void *data)
{
  return 0;
}    

static int led_drv_ver_input(char *buf, char **start, off_t offset, int count, 
  int *eof, void *data)
{
  int len=0, i;
  struct gpio_input_state input_gpio;
  
  for (i = 0; i < LED_MAX_GPIO_INPUTS; i++) {
    if (led_hal_get_input_status(i, &input_gpio) == 0 ) 
    {
  	  len += sprintf(buf + len, "\nGPIO %d: %d", input_gpio.gpio_num, input_gpio.gpio_state);
    }  
  }
    
  if (len <= offset + count)
    *eof = 1;
   
  *start = buf + offset;
  len -= offset;
    
  if (len > count)
    len = count;
    
  if (len < 0)
    len = 0;  
      
	return len;                                                                                
}                                                                                              

int __init led_drv_init(void)
{
  led_dev = kmalloc(sizeof(struct led_device), GFP_KERNEL);
  if (!led_dev)
    return -ENOMEM;

  /* register LED device */
  devfs_register(NULL, "led_drv", DEVFS_FL_AUTO_DEVNUM, 0, 0,
      S_IFCHR | S_IRUGO | S_IWUGO, &led_drv_fops,NULL);

  led_dev->status_change = 0;
  led_dev->lock = SPIN_LOCK_UNLOCKED;
  init_waitqueue_head(&led_dev->input_status_queue);
  
  led_hal_init(gpio_off_state, (void *)led_dev, 0);

  /* create proc entry */
  create_proc_read_entry("avalanche/led_drv_cfg", 0, NULL, led_drv_cfg_info, NULL);
  create_proc_read_entry("avalanche/led_drv_ver", 0, NULL, led_drv_ver_info, NULL);
  create_proc_read_entry("avalanche/led_drv_input", 0, NULL, led_drv_ver_input, NULL);
  return 0;
}

void __exit led_drv_exit(void)
{
  led_hal_exit();
    
  remove_proc_entry("avalanche/led_drv_cfg", NULL);
  remove_proc_entry("avalanche/led_drv_ver", NULL);
  remove_proc_entry("avalanche/led_drv_input", NULL);

  kfree(led_dev);
}

module_init(led_drv_init);
module_exit(led_drv_exit);

EXPORT_SYMBOL_NOVERS(led_hal_register);
EXPORT_SYMBOL_NOVERS(led_hal_action);
EXPORT_SYMBOL_NOVERS(led_hal_deregister);
