/******************************************************************************
 * FILE PURPOSE:     - LED HAL module source
 ******************************************************************************
 * FILE NAME:     led_hal.c
 *
 * DESCRIPTION:    LED HAL API's source.
 *
 *
 * (C) Copyright 2004, Texas Instruments, Inc
 *******************************************************************************/
#include "led_platform.h"

/* include for gpio APIs */
#include "led_hal_drv.h"

#define LED_MAGIC    0xED0A243C
#define LED_MAX_LEDS   32

/* Default timeout for INPUT GPIO Timer 100 ms */
#define DEFAULT_INPUT_TIMEOUT 100

#define LED_HAL_SKIP_SANITY_CHECKS

/****** Driver internal types *******/
struct led_cur_state
{
  LED_MODE_T   led_status;    /* LED current status On/Off/Blink */

  LED_MODE_T   mode;          /* Requested mode: On/Off/Blink */
  unsigned int param1;			  /* Parameter 1 (optional) */
  unsigned int param2;			  /* Parameter 2 (optional) */

  unsigned int timer_running; /* Timer running flag */
  void *os_timer;	    				/* Function timer pointer */
};

/* This structure holds information about a single LED and its state */
struct led_physical
{
  unsigned int led_magic;
  struct led_funcs *led_funcs;      	      /* LED functions for non-gpio LED */
  struct led_cur_state *led_current_state;  /* LED current state */  
  unsigned int led_gpio;
};

struct led_map
{
  unsigned int led_map[LED_MAX_COLORS];
  unsigned int led_mask;
  unsigned int color_index;
};

struct led_hal_device
{
  unsigned int gpio_offstate_map;
  struct led_physical **led_general_array;
  struct gpio_input_state **gpio_general_array;  
  unsigned int gpio_lock;
  void *input_os_timer;  
  int input_timeout;
  void *drv_data;  
};

/****** Driver internal data *******/
static struct led_hal_device *led_hal_dev;

/****** Driver internal functions ********/
static void led_init(int param);
static void led_on(int param);
static void led_off(int param);
static void led_timer_func(int arg);
static void *led_malloc(int n);
static void gpio_timer_func(int arg);
void led_timer_func2( void *t, int arg); /* This is not static because of compatiblity with vxworks */

/* Notify function, to be called whenever there is a change in the state of Input GPIO */
void led_drv_notify_state_change(void *drv_data);

/* Registers a LED, returns LED handle or negative value on error */
int led_hal_user_register(const struct led_gpio_object *led)
{
  struct led_object led_obj;
  int i;
  
  if (!led)
    return -EINVAL;
  
  led_obj.led_funcs = NULL;
  led_obj.num_of_gpios = led->num_of_gpios;
  
  for (i = 0; i < LED_MAX_BITS; i++)
    led_obj.led_gpios[i] = led->led_gpios[i];
    
  return (led_hal_register(&led_obj));  
}

int led_hal_input_register(const struct input_gpio_object *gpio_obj)
{
  int i;
   
  if (!led_hal_dev->gpio_general_array || !led_hal_dev->led_general_array)
    return -ENOMEM;
  
  /* Look if someone already has this LED */
  for (i = 0; i < LED_MAX_LEDS; i++) 
  {
    if ((led_hal_dev->led_general_array[i]) && (led_hal_dev->led_general_array[i]->led_gpio == gpio_obj->input_gpio))
      return -EBUSY;
  }
  for (i = 0; i < LED_MAX_GPIO_INPUTS; i++) 
  {
    if ((led_hal_dev->gpio_general_array[i]) && (led_hal_dev->gpio_general_array[i]->gpio_num == gpio_obj->input_gpio))
      return -EBUSY;
  }

  PAL_osProtectEntry( PAL_OSPROTECT_INTERRUPT, &led_hal_dev->gpio_lock );

  for (i = 0; i < LED_MAX_GPIO_INPUTS; i++) {
    if (led_hal_dev->gpio_general_array[i] == NULL)
    {
      struct gpio_input_state *gpio = NULL;
      led_hal_dev->gpio_general_array[i] = (struct gpio_input_state *)led_malloc(sizeof(struct gpio_input_state));
      gpio = led_hal_dev->gpio_general_array[i];
      if (!gpio)
      {
        PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
        return -ENOMEM;
      }  
#ifdef LED_HAL_DEBUG      
      log_msg("Handle %d returned to input GPIO %d\n", i, gpio->gpio_num);
#endif      
      /* Initialize GPIO pin */
      gpio->gpio_num = gpio_obj->input_gpio;
      gpio->gpio_state = 0;
      gpio->state_change = 0;
      gpio->invert = gpio_obj->invert;

      PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
      
      PAL_sysGpioCtrl(gpio->gpio_num, GPIO_PIN, GPIO_INPUT_PIN);
      if ((i == 0) && (led_hal_dev->input_os_timer))
      {
        OS_TIMER_ADD(led_timer_func2, led_hal_dev->input_os_timer, led_hal_dev->input_timeout, 0); 
      }
      return i;
    }
  }

  PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
  return -ENOMEM;  
}

/* Deregisters a registered input GPIO */
int led_hal_input_deregister(int gpio_handle)
{
  if (gpio_handle > LED_MAX_GPIO_INPUTS)
    return -EINVAL;

  if (!led_hal_dev->gpio_general_array[gpio_handle])
    return -EINVAL;

#ifdef LED_HAL_DEBUG      
  log_msg("Freeing handle %d", gpio_handle);
#endif      

  PAL_osProtectEntry( PAL_OSPROTECT_INTERRUPT, &led_hal_dev->gpio_lock );

  PAL_osMemFree(0,(void *)led_hal_dev->gpio_general_array[gpio_handle],sizeof(struct gpio_input_state));
  led_hal_dev->gpio_general_array[gpio_handle] = 0;

  PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
  return 0;  
}

int led_hal_get_input_status(int index, struct gpio_input_state *gpio)
{
  if (!led_hal_dev->gpio_general_array)
    return -EINVAL;

  if (!gpio)
    return -EINVAL;
  
  if (index > LED_MAX_GPIO_INPUTS)
    return -EINVAL;
    
  if (led_hal_dev->gpio_general_array[index] == NULL )
    return -ENXIO;
  else
  {
    gpio->gpio_num = led_hal_dev->gpio_general_array[index]->gpio_num;
    gpio->gpio_state = led_hal_dev->gpio_general_array[index]->gpio_state;
    gpio->state_change = led_hal_dev->gpio_general_array[index]->state_change;
  }  
  return 0;
}

/* Registers a LED, returns LED ID */
int led_hal_register(const struct led_object *led)
{
  struct led_physical *led_phy;
  unsigned int num_of_gpios;
  int i;
  
#ifdef LED_HAL_DEBUG      
  log_msg( "led_hal_register\n" );
#endif      
  
  if (!led)
    return -EINVAL;   

  if (!led_hal_dev->led_general_array || !led_hal_dev->gpio_general_array)
    return -ENOMEM;

  /* Get number of GPIO pins */
  if (led->num_of_gpios > LED_MAX_BITS || led->num_of_gpios == 0) 
    num_of_gpios = 1;
  else
    num_of_gpios = led->num_of_gpios;
  
  /* Currently - Only 2 GPIOs supported */
  if (num_of_gpios > 2)
    return -EINVAL;

#ifdef LED_HAL_DEBUG      
  log_msg( "Num of GPIOs: %d\n", num_of_gpios );
#endif      

  /* Check if user supplied LED functions but forgot to supply on/off functions */
  if (led->led_funcs)
  {
    if(!led->led_funcs->led_on || !led->led_funcs->led_off) 
      return -EINVAL; 
  }
  else
  {
     /* Look for invalid GPIOs */
     for (i = 0; i < num_of_gpios; i++) { 
       if (led->led_gpios[i] > 31)
         return -EINVAL;
#ifdef LED_HAL_DEBUG      
       log_msg("GPIO: %d\n", led->led_gpios[i]);
#endif
     }

     /* Look if someone already has this LED */
     for (i = 0; i < LED_MAX_LEDS; i++)
     {
       if ((led_hal_dev->led_general_array[i]) && 
           ((led_hal_dev->led_general_array[i]->led_gpio == led->led_gpios[0]) ||
            ((led->num_of_gpios > 1) && (led->led_gpios[1] == led_hal_dev->led_general_array[i]->led_gpio))))
       {
         return i;
       }    
     }
     for (i = 0; i < LED_MAX_GPIO_INPUTS; i++) 
     {
       if ((led_hal_dev->gpio_general_array[i]) && 
           ((led_hal_dev->gpio_general_array[i]->gpio_num == led->led_gpios[0]) || 
            ((led->num_of_gpios > 1) && (led_hal_dev->gpio_general_array[i]->gpio_num == led->led_gpios[1])))) 
       {
         return -EBUSY;
       }
     }
  }
  
  /* Allocate memory */
  led_phy = led_malloc(sizeof(struct led_physical));
  if (!led_phy)
    return -ENOMEM;

  led_phy->led_magic = LED_MAGIC;  
  led_phy->led_funcs = led_malloc(sizeof(struct led_funcs));
  if (!led_phy->led_funcs)
  {
    PAL_osMemFree(0,led_phy,sizeof(struct led_physical));
    return -ENOMEM;
  }

  led_phy->led_current_state = led_malloc(sizeof(struct led_cur_state));
  if (!led_phy->led_current_state)
  {
    PAL_osMemFree(0,led_phy->led_funcs,sizeof(struct led_funcs));
    PAL_osMemFree(0,led_phy,sizeof(struct led_physical));
    return -ENOMEM;
  }  
  
  led_phy->led_current_state->os_timer = os_timer_init(led_timer_func);
  if (!led_phy->led_current_state->os_timer)
  {
    PAL_osMemFree(0,led_phy->led_funcs,sizeof(struct led_funcs));
    PAL_osMemFree(0,led_phy->led_current_state,sizeof(struct led_cur_state));
    PAL_osMemFree(0,led_phy,sizeof(struct led_physical));
    return -ENOMEM;   
  }
  
  /* Initialize LED functions or GPIO */
  if (led->led_funcs)
  {
    PAL_osMemCopy(led_phy->led_funcs, led->led_funcs, sizeof(struct led_funcs));    
    led_phy->led_gpio = 0xFF;
  }  
  else
  {   
    struct led_map *led_info;
    
    led_info = led_malloc(sizeof(struct led_map));
    if (!led_info)
    {
      PAL_osMemFree(0,led_phy->led_funcs, sizeof(struct led_funcs));
      PAL_osMemFree(0,led_phy->led_current_state, sizeof(struct led_cur_state));
      os_timer_delete(led_phy->led_current_state->os_timer);
      PAL_osMemFree(0,led_phy, sizeof(struct led_physical));
      return -ENOMEM;   
    }
    led_phy->led_gpio = led->led_gpios[0];
    
    /* Calculate LED's mask */
    led_info->led_mask = 0;
    for (i = 0; i < num_of_gpios; i++)  
    {
      led_info->led_mask |= (1 << led->led_gpios[i]); 
    }  
#ifdef LED_HAL_DEBUG      
    log_msg("LED Mask: 0x%x\n", led_info->led_mask);
#endif         
    /* Get LED's bitmap */
    if (led->num_of_gpios == 1)
    {
      led_info->led_map[0] = led_info->led_mask;
    }
    else
    {
      /* Two GPIOs */
      led_info->led_map[0] = (1 << led->led_gpios[1]);
      led_info->led_map[1] = (1 << led->led_gpios[0]);
      led_info->led_map[2] = (1 << led->led_gpios[0]) | ( 1 << led->led_gpios[1]);
    }
          
    /* Init functions */
    led_phy->led_funcs->param = (unsigned int)led_info;    
    led_phy->led_funcs->led_init = led_init;
    led_phy->led_funcs->led_off = led_off;
    led_phy->led_funcs->led_on = led_on;
  }
    
  /* Initialize LED */
  if (led_phy->led_funcs->led_init)
    led_phy->led_funcs->led_init(led_phy->led_funcs->param);

  PAL_osProtectEntry( PAL_OSPROTECT_INTERRUPT, &led_hal_dev->gpio_lock );

  for (i = 0; i < LED_MAX_LEDS; i++)
  {
    if (led_hal_dev->led_general_array[i] == NULL)
    {
      led_hal_dev->led_general_array[i] = led_phy;
#ifdef LED_HAL_DEBUG      
      log_msg("Handle %d returned to LED GPIO %d\n", i, led_phy->led_gpio);
#endif      
      PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
      return i;
    }  
  }  

  PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
  
  PAL_osMemFree(0,led_phy->led_funcs,sizeof(struct led_funcs));
  PAL_osMemFree(0,led_phy->led_current_state,sizeof(struct led_cur_state));
  os_timer_delete(led_phy->led_current_state->os_timer);
  PAL_osMemFree(0,led_phy,sizeof(struct led_physical));
  return -ENOMEM;  
}

/* Deregisters a registered LED */
int led_hal_deregister(int led_handle)
{
  struct led_physical *led_phy;
  
  if (led_handle > LED_MAX_LEDS)
    return -EINVAL;
  else
    led_phy = led_hal_dev->led_general_array[led_handle];
  
  if (!led_phy)
    return -EINVAL;   

  if (led_phy->led_magic != LED_MAGIC)
    return -EINVAL;    

#ifdef LED_HAL_DEBUG      
   log_msg( "Freeing handle %d", led_handle);
#endif      

  PAL_osProtectEntry( PAL_OSPROTECT_INTERRUPT, &led_hal_dev->gpio_lock );

  /* Delete timer */
  if (led_phy->led_current_state->os_timer)
  {
     os_timer_delete(led_phy->led_current_state->os_timer);
     led_phy->led_current_state->os_timer = 0;
  }

  /* Initialize LED */
  led_phy->led_current_state->mode = LED_MODE_OFF;
  led_phy->led_current_state->led_status = LED_MODE_OFF;
  
  if (led_phy->led_funcs->led_init)
    led_phy->led_funcs->led_init(led_phy->led_funcs->param); 

  /* Free memory */
  PAL_osMemFree(0,led_phy->led_funcs,sizeof(struct led_funcs));
  PAL_osMemFree(0,led_phy->led_current_state,sizeof(struct led_cur_state));
  PAL_osMemFree(0,led_phy,sizeof(struct led_physical));
  led_hal_dev->led_general_array[led_handle] = 0;

  PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
  return 0;  
}

/* Perform an action on a LED */
int led_hal_action(const struct led_action *led_action)
{
  struct led_physical *led_phy;
  struct led_cur_state *led_current_state;
  
#ifdef LED_HAL_DEBUG      
  log_msg( "led_hal_action\n" );
#endif      
    
#ifndef LED_HAL_SKIP_SANITY_CHECKS
  /* Sanity check */
  if (!led_action)
    return -EINVAL;

  if (led_action->led_handle < 0 || led_action->led_handle >= LED_MAX_LEDS)
    return -EINVAL;
#endif
    
  PAL_osProtectEntry( PAL_OSPROTECT_INTERRUPT, &led_hal_dev->gpio_lock );

  led_phy = (struct led_physical *)led_hal_dev->led_general_array[led_action->led_handle];
  
#ifndef LED_HAL_SKIP_SANITY_CHECKS
  if (!led_phy)
    return -EINVAL;

  if (led_phy->led_magic != LED_MAGIC)
    return -EINVAL;  
#endif
  
  led_current_state = led_phy->led_current_state; 
  
  /* Check if timer is running...*/
  if (led_current_state->timer_running)
  {
    /* Check if module requested the same action */
    if ( led_current_state->mode == led_action->mode )
    {
      PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );   

      /* Don't do anything - Resolve the over blink problem */
      return 0;
    }
    
    /* New action to be performed on the LED */
    led_current_state->timer_running = 0;
    os_timer_delete(led_current_state->os_timer);
  }

  /* Init color index for NON-GPIO LEDs */
  if ( led_phy->led_gpio != 0xFF )
  {
     struct led_map *led_info = (struct led_map *) led_phy->led_funcs->param;

     if( led_action->color_index )
        led_info->color_index = led_action->color_index - 1;
     else   
        led_info->color_index = 0;
     
#ifdef LED_HAL_DEBUG      
     log_msg( "Color index: %d\n", led_info->color_index );
#endif             
  }
    
  /* Init mode/status */
  led_current_state->mode = led_action->mode;  
  led_current_state->led_status = led_action->mode;  

  /* Activate an action */
  switch (led_action->mode)
  {
    unsigned int param1;
    
    case LED_MODE_OFF:
      PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
      led_phy->led_funcs->led_off(led_phy->led_funcs->param);
      return 0;
    
    case LED_MODE_ON:
      PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
      led_phy->led_funcs->led_on(led_phy->led_funcs->param);
      return 0;  

    case LED_MODE_ONE_SHOT_OFF:
    case LED_MODE_ONE_SHOT_ON:
#ifndef LED_HAL_SKIP_SANITY_CHECKS
      if(led_action->param1 < 50)
      {
        PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
        return -EINVAL;
      }
#endif        
      /* Add timer to time based actions */
      led_current_state->timer_running = 1;
      led_current_state->param1 = led_action->param1;
    
      param1 = ( led_current_state->param1 << 3 ) / 10 ;

      OS_TIMER_ADD(led_timer_func2, led_current_state->os_timer,param1,(int)led_phy);
        
      PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );

      if (led_action->mode == LED_MODE_ONE_SHOT_OFF)
        led_phy->led_funcs->led_off(led_phy->led_funcs->param);
      else
        led_phy->led_funcs->led_on(led_phy->led_funcs->param);
    
      break;
      
    case LED_MODE_BLINK:
#ifndef LED_HAL_SKIP_SANITY_CHECKS
      if(led_action->param1 < 50)
      {
        PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
        return -EINVAL;
      }
        
      if (led_action->param2 < 50)
      {
        PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
        return -EINVAL;
      }       
#endif

      /* Add timer to time based actions */
      led_current_state->timer_running = 1;
      led_current_state->param1 = led_action->param1;
      led_current_state->param2 = led_action->param2;
      param1 = led_current_state->param1;  
        
      OS_TIMER_ADD(led_timer_func2, led_current_state->os_timer,param1,(int)led_phy);
        
      PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );

      led_phy->led_funcs->led_on(led_phy->led_funcs->param);
      break;

    default:
      PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
      return -EINVAL;
  } 
  return 0;
}

/* HAL control function */
int led_hal_control(int input_timeout)
{
  if ((input_timeout < 10) || (input_timeout > 500))
    return -1;
  led_hal_dev->input_timeout = input_timeout;
  
  OS_TIMER_ADD(led_timer_func2, led_hal_dev->input_os_timer, led_hal_dev->input_timeout, 0); 
  
  return 0;
}

/* Module init function */
int led_hal_init(unsigned int gpio_off_value, void *drv_data, unsigned int args )
{
  led_hal_dev = led_malloc(sizeof(struct led_hal_device));
  if (!led_hal_dev)
    return -ENOMEM;

  led_hal_dev->gpio_offstate_map = gpio_off_value;
  led_hal_dev->drv_data = drv_data;
  led_hal_dev->input_timeout = DEFAULT_INPUT_TIMEOUT;

  led_hal_dev->input_os_timer = os_timer_init(gpio_timer_func);
  if (!led_hal_dev->input_os_timer)
  {
    PAL_osMemFree(0,led_hal_dev,sizeof(struct led_hal_device));
    return -ENOMEM;
  }

  led_hal_dev->led_general_array = led_malloc(LED_MAX_LEDS * sizeof(struct led_physical *));
  if (!led_hal_dev->led_general_array)
  {
    PAL_osMemFree(0,led_hal_dev,sizeof(struct led_hal_device));
    return -ENOMEM;
  }
  led_hal_dev->gpio_general_array = led_malloc(LED_MAX_GPIO_INPUTS * sizeof(struct gpio_input_state *));
  if (!led_hal_dev->gpio_general_array)
  {
    PAL_osMemFree(0,led_hal_dev->led_general_array,sizeof(struct gpio_input_state));
    PAL_osMemFree(0,led_hal_dev,sizeof(struct led_hal_device));
    return -ENOMEM;
  }
#ifdef LED_HAL_DEBUG      
  log_msg("\n\n\n******** LED HAL INIT **********\n\n\n");
#endif  
  led_hal_dev->gpio_lock = 0;

  return 0;
}

/* Module exit function */
int led_hal_exit(void)
{
  int i;
  
  if (led_hal_dev->led_general_array)
  {
    /* Deregister all LEDs */
    for (i = 0; i < LED_MAX_LEDS; i++)
      led_hal_deregister(i);
    PAL_osMemFree(0,led_hal_dev->led_general_array,sizeof(struct gpio_input_state));    
  }  
  if (led_hal_dev->gpio_general_array)
  {
    /* Deregister all GPIOs */
    for (i = 0; i < LED_MAX_LEDS; i++)
      led_hal_input_deregister(i);
    PAL_osMemFree(0,led_hal_dev->gpio_general_array,sizeof(struct gpio_input_state));
  }

  /* Delete timer */
  if (led_hal_dev->input_os_timer)
  {
     os_timer_delete(led_hal_dev->input_os_timer);
     led_hal_dev->input_os_timer = 0;
  }
  return 0;
}

/********* Internal functions **********/

/* Function to initialize a GPIO LED */
static void led_init(int param)
{
  struct led_map *led_info = (struct led_map *)param;
  unsigned int value, i;

  if (!led_info)
    return;
#ifdef LED_HAL_DEBUG      
  log_msg( "---LED INIT\n" );
#endif      

  for (i = 0; i < 32; i++)
  {
    if (led_info->led_mask & (1 << i))
    {
#ifdef LED_HAL_DEBUG      
      log_msg("Init GPIO: %d, mask: 0x%x\n", i, led_info->led_mask);
#endif
      /* Initialize GPIO pin */
      PAL_sysGpioCtrl(i, GPIO_PIN, GPIO_OUTPUT_PIN);    
      value = (led_hal_dev->gpio_offstate_map & (led_info->led_mask & (1 << i))) ? 1 : 0; 
      PAL_sysGpioOutBit(i, value);
    }
  }        
}

/* Function to turn GPIO LED on */
static void led_on(int param)
{
  struct led_map *led_info = (struct led_map *)param;
  
#ifndef LED_HAL_SKIP_SANITY_CHECKS
  if (!led_info)
    return;
#endif
#ifdef LED_HAL_DEBUG     
  log_msg("---LED ON\n");
  log_msg("Map: 0x%x, mask: 0x%x\n", ~((~led_hal_dev->gpio_offstate_map) ^ led_info->led_map[led_info->color_index]) & led_info->led_mask, led_info->led_mask);
#endif
  PAL_sysGpioOutValue(~((~led_hal_dev->gpio_offstate_map) ^ led_info->led_map[led_info->color_index]) & led_info->led_mask, led_info->led_mask, 0);
}

/* Function to turn GPIO LED off */
static void led_off(int param)
{
  struct led_map *led_info = (struct led_map *)param;

#ifndef LED_HAL_SKIP_SANITY_CHECKS
  if (!led_info)
    return;
#endif
#ifdef LED_HAL_DEBUG      
  log_msg("---LED OFF\n");
  log_msg("Map: 0x%x, mask: 0x%x\n", led_hal_dev->gpio_offstate_map & led_info->led_map[led_info->color_index], led_info->led_mask);
#endif
  PAL_sysGpioOutValue((led_hal_dev->gpio_offstate_map & led_info->led_mask), led_info->led_mask, 0);
}

static void gpio_timer_func(int arg)
{
  struct gpio_input_state *gpio;
  int state_change = 0, state, i;
  
  for (i = 0; i < LED_MAX_GPIO_INPUTS; i++) 
  {
    gpio = led_hal_dev->gpio_general_array[i];
    if (gpio == NULL)
      continue;
    state = PAL_sysGpioInBit(gpio->gpio_num) ^ gpio->invert;    
    if (gpio->gpio_state != state)
    {
      gpio->gpio_state = state;
      gpio->state_change = 1;
      state_change++;
    }
  }

  if (state_change)
    led_drv_notify_state_change(led_hal_dev->drv_data);

  OS_TIMER_ADD(led_timer_func2, led_hal_dev->input_os_timer, led_hal_dev->input_timeout, 0); 
}    

/* LED timer function with two arguments */
void led_timer_func2( void *t, int arg)
{
  led_timer_func( arg );
}

/* LED timer function */
static void led_timer_func(int arg)
{   
  struct led_funcs *led_funcs;
  struct led_cur_state *led_current_state;

#ifndef LED_HAL_SKIP_SANITY_CHECKS
  if (!arg)
    return;
#endif
    
  PAL_osProtectEntry( PAL_OSPROTECT_INTERRUPT, &led_hal_dev->gpio_lock );  
    
  led_funcs = ((struct led_physical *)arg)->led_funcs;
  led_current_state = ((struct led_physical *)arg)->led_current_state;

  if (led_current_state->timer_running == 0)
  {
     PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
     return;
  }
  
  switch (led_current_state->mode)
  {
    case LED_MODE_BLINK:
#ifdef LED_HAL_DEBUG      
      log_msg( "---LED BLINK\n" );
#endif      
      if (led_current_state->timer_running == 1)
      {
        led_current_state->timer_running = 2;
        led_funcs->led_off(led_funcs->param);
        OS_TIMER_ADD(led_timer_func2, led_current_state->os_timer, led_current_state->param2, (int)arg); 
        
      }
      else
      {
        led_current_state->timer_running = 1;
        led_funcs->led_on(led_funcs->param);

        OS_TIMER_ADD(led_timer_func2, led_current_state->os_timer, led_current_state->param1, (int)arg);  
      }  
      break;
      
    case LED_MODE_ONE_SHOT_OFF:
      if (led_current_state->timer_running == 1)
      {
        led_current_state->timer_running = 2;
        led_current_state->led_status = LED_MODE_ON;      
        led_funcs->led_on(led_funcs->param);
        OS_TIMER_ADD(led_timer_func2, led_current_state->os_timer, ( ( led_current_state->param1 << 1 ) / 10 ), (int)arg); 
      }
      else
      {
        led_current_state->timer_running = 0;
      }
      break;  
        
    case LED_MODE_ONE_SHOT_ON:
      if (led_current_state->timer_running == 1)
      {
        led_current_state->timer_running = 2;
        led_current_state->led_status = LED_MODE_OFF;      
        led_funcs->led_off(led_funcs->param);
        OS_TIMER_ADD(led_timer_func2, led_current_state->os_timer, ( ( led_current_state->param1 << 1 ) / 10 ), (int)arg); 
      }
      else
      {
        led_current_state->timer_running = 0;
      }
      break;
          
    default:
      break;
  }

  PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_hal_dev->gpio_lock );
}  
  
/* Allocate memory and clear it */
static void *led_malloc(int n)
{
  void *p = NULL;
  PAL_Result ret;

  ret = PAL_osMemAlloc(0,n,4,&p);
  if (ret == PAL_SOK)
      PAL_osMemSet(p, 0, n);
  return p;
}

