/******************************************************************************    
 * FILE PURPOSE:     - LED manager module Source                                       
 ******************************************************************************    
 * FILE NAME:     led_manager.c                                                        
 *                                                                                 
 * DESCRIPTION:   Linux LED manager character driver implementation                                                  
 *                                                                                 
 *                                                                                 
 * (C) Copyright 2004, Texas Instruments, Inc                                      
 *******************************************************************************/  
                                                                                   
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/delay.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/wait.h>
#include <linux/proc_fs.h>
#include <asm/avalanche/generic/pformCfg.h>
#include <asm/avalanche/generic/pal.h>
#include "led_platform.h"
#include "led_drv_user.h"
#include "led_hal_drv.h"
#include "led_manager.h"
#include "ledcfg.h"

/*=============================================================================
   Defines, data types, globals, and function prototypes
  =============================================================================*/

#define LED_MANAGER_MAGIC_NUMBER         0xAC703BBC          /* Magic number */
#define LED_MANAGER_SKIP_SANITY_CHECKS      /* Skip runtime parameters check */
/* #define LED_MANAGER_DEBUG */

#define LED_MANAGER_MESSAGE_PREFIX       "led_manager: "

struct led_state
{
  int handle;                      /* LED object handle */
  struct led_object *object;       /* LED object */
  struct led_action *action;       /* State action */
};

struct led_module
{
  unsigned int magic;              /* Magic number */
  char name[ MODULE_NAME_LEN ];    /* Module name */
  int instance;                    /* Module instance */
  struct led_state *states[ MAX_STATE_ENTRIES ];  /* Pointer to array of states */
  struct led_module *next;         /* Pointer to the next module */
};

struct led_module *led_module_head = NULL;

/* Static functions */
static int led_manager_configure( struct led_config *led_cfg );

/************************************************************************
 *  Function: led_manager_register_module 
 *
 *  Purpose:  Registers a module in the LED manager database
 *                                                                     
 *  Inputs: module_name - A string with the module name
 *		    instance - Instance number
 *
 *  Outputs: Handle to entity on success, or NULL if failed.
 *                                                                     
 ************************************************************************/
void *led_manager_register_module( const char *module_name, int instance )
{
  struct led_module *module = led_module_head;
  
  /* Check parameters */
  if( !module_name )
     return NULL;

  while ( module )
  {
     /* Find the module in the list */
     if( !strcmp( module_name, module->name ) && ( instance == module->instance ) )
     {
#ifdef LED_MANAGER_DEBUG
        printk( "%sRegistered %s, instance %d\n", LED_MANAGER_MESSAGE_PREFIX, module->name, module->instance );
#endif
        break;
     }
     
     module = module->next;
  }     
  
  /* Module not registered, register a NULL device */
  if ( !module )
  {
     struct led_config ledcfg;
    
     memset( &ledcfg, 0, sizeof( struct led_config ) );
    
     strncpy( ledcfg.name, module_name, MODULE_NAME_LEN );
     ledcfg.gpio_num = 0;
     ledcfg.instance = instance;
        
     if( led_manager_configure( &ledcfg ) < 0 )
        return NULL;
  
     return led_manager_register_module( module_name, instance );
  }
 
  return module;
}

/************************************************************************
 *  Function: led_manager_unregister_module 
 *
 *  Purpose:  Unregisters a module from the LED manager database
 *                                                                     
 *  Inputs: module_handle - Handle returned by the register function
 *
 *  Outputs: 0 on success.
 *           -EINVAL if module not found  
 *                                                                     
 ************************************************************************/
int led_manager_unregister_module( const void *module_handle )
{
  struct led_module *module = ( struct led_module *)module_handle;
  struct led_module *search = led_module_head;
  unsigned int led_manager_lock;
  int i, found = 0;

  if( !module )
     return -EINVAL;

  PAL_osProtectEntry( PAL_OSPROTECT_INTERRUPT, &led_manager_lock );  

  if ( module == led_module_head )
  {
     found = 1;
  }
  else
  {   
     /* Find module in list */
     while ( search )
     {
        if ( search->next == module )
        {       
           found = 1;
           break;
        }
        
        search = search->next;
     }
  }
  
  if ( found )
  {      
     for ( i = 0; i < MAX_STATE_ENTRIES; i++ )
     {    
        if ( module->states[ i ] )
        {
           if ( module->states[ i ]->object->led_funcs )
           {
           /* Unregister the LED associated with the state */
              if ( !led_hal_deregister( module->states[ i ]->handle ) )             
              {
                 /* Clear callbacks */
                 memset( module->states[ i ]->object->led_funcs, 0, sizeof( struct led_funcs ) );
              }
           }
        }
     }

#ifdef LED_MANAGER_DEBUG
     printk( "%sUnregistered %s, instance %d\n", LED_MANAGER_MESSAGE_PREFIX, module->name, module->instance );
#endif
  }     
  
  PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_manager_lock );  
  
  if( found )
    return 0;
  
  return -EINVAL;  
}

/************************************************************************
 *  Function: led_manager_install_callbacks 
 *
 *  Purpose:  Install LED callback functions (required for Non-GPIO LEDs)
 *                                                                     
 *  Inputs: module_handle - Handle returned by the register function
 *          led_id - ID of the LED (0,1...)
 *          led_funcs - Structure with all callback functions
 *
 *  Outputs: 0 on success.
 *           -EINVAL on failure  
 *                                                                     
 ************************************************************************/
int led_manager_install_callbacks( void *module_handle, int led_id, struct led_funcs *funcs )
{
  struct led_module *module = ( struct led_module *)module_handle;
  int ret = -1, i, reg = 0;
  unsigned int led_manager_lock;
  
  /* Check parameters */
  if( !module_handle )
     return -EINVAL;
  
  /* Check parameters */
  if( module->magic != LED_MANAGER_MAGIC_NUMBER )
     return -EINVAL;

  if( !funcs )
     return -EINVAL; 
     
  PAL_osProtectEntry( PAL_OSPROTECT_INTERRUPT, &led_manager_lock );  
    
  for( i = 0; i < MAX_STATE_ENTRIES; i++ )
  {
     if( !module->states[ i ] )
        continue;
        
     /* Find the LED in the list */
     if ( module->states[ i ]->object->led_gpios[ 0 ] == 0xFF - led_id )
     {       
        /* Register LED only once */
        if ( reg == 0 )
        {
           reg = 1;
           
           memcpy( module->states[ i ]->object->led_funcs, funcs, sizeof( struct led_funcs ) );
           
           /* Register the Non-GPIO LED */
           ret = led_hal_register( module->states[ i ]->object );
        }
        
        if ( ret < 0 )
        {
           /* Free all states with the problematic functionallity */
           kfree( module->states[ i ]->object );
           kfree( module->states[ i ]->action );
           kfree( module->states[ i ] );
           module->states[ i ] = NULL;
        }
        else
        {
           /* Save the handle */
           module->states[ i ]->handle = ret;
        
           /* Save the handle in the action structure for the LED HAL */
           module->states[ i ]->action->led_handle = ret;
        }
     }
  }     

  PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_manager_lock );  
  
  /* Probably there is no such LED ID */
  if( reg == 0 )
     return -EINVAL;

  if( ret < 0 )
     return ret;
  
  return 0;   
}

/************************************************************************
 *  Function: led_manager_led_action 
 *
 *  Purpose:  Triggers an action on an LED
 *                                                                     
 *  Inputs: module_handle - Handle returned by the register function
 *          state_id - ID of the state (from configuration file)
 *
 *  Outputs: 0 on success.
 *           -EINVAL on failure  
 *                                                                     
 ************************************************************************/
int led_manager_led_action( void *module_handle, int state_id )
{
  struct led_module *module = ( struct led_module *)module_handle;

#ifndef LED_MANAGER_SKIP_SANITY_CHECKS

  /* Check parameters */
  if ( !module_handle )
  {
#ifdef LED_MANAGER_DEBUG
     printk( "%sModule handle is NULL!\n", LED_MANAGER_MESSAGE_PREFIX );
#endif 
     return -EINVAL;
  }
  
  /* Check parameters */
  if( module->magic != LED_MANAGER_MAGIC_NUMBER )
  {
#ifdef LED_MANAGER_DEBUG
     printk( "%sBad magic number!\n", LED_MANAGER_MESSAGE_PREFIX );
#endif 
     return -EINVAL;
  }

  if( state_id > MAX_STATE_ENTRIES ) 
  {
#ifdef LED_MANAGER_DEBUG
     printk( "%sInvalid state ID %d! (Module: %s)\n", LED_MANAGER_MESSAGE_PREFIX, state_id, module->name );
#endif 
     return -EINVAL;
  }

#endif  /* LED_MANAGER_SKIP_SANITY_CHECKS */
  
  if ( module->states[ state_id ] ) 
  {
    /* Perform action */
    return led_hal_action( module->states[ state_id ]->action );
  }
  
  return -EINVAL;    
}


/************************************************************************
 *  Function: led_manager_configure 
 *
 *  Purpose:  Configure a module and its states.
 *                                                                     
 *  Inputs: led_cfg - LED configuration structure
 *
 *  Outputs: 0 on success.
 *           -EINVAL on failure
 *           -ENOMEM on memory allocation failure
 *           -EBUSY on multiple configurations of the same state  
 *                                                                     
 ************************************************************************/
static int led_manager_configure( struct led_config *led_cfg )
{
  struct led_module *module;
  struct led_state *state;
  int ret = 0;
  unsigned int led_manager_lock;
   
  module = led_module_head;
  
  /* Look through list to see if we have this module */
  while( module )
  {
     if ( !strcmp( led_cfg->name, module->name ) && ( led_cfg->instance == module->instance ) )
        break;
     
     module = module->next;
  }
  
  /* Did we find such module? */
  if ( !module )
  {          
#ifdef LED_MANAGER_DEBUG
     printk( "%sRegistering module %s\n", LED_MANAGER_MESSAGE_PREFIX, led_cfg->name );
#endif
     /* No, make a new node */
     if ( ( module = kmalloc( sizeof( struct led_module ), GFP_KERNEL ) ) == NULL )
       return -ENOMEM;

     memset( module, 0, sizeof( struct led_module ) );
     
     /* Init parameters */
     module->magic = LED_MANAGER_MAGIC_NUMBER;
     strcpy( module->name, led_cfg->name );
     module->instance = led_cfg->instance;
     
     PAL_osProtectEntry( PAL_OSPROTECT_INTERRUPT, &led_manager_lock );  

     /* Add new module to the list */
     module->next = led_module_head;
     led_module_head = module;
     
     PAL_osProtectExit( PAL_OSPROTECT_INTERRUPT, led_manager_lock );           
  }              
  
  /* Check for NULL device - return OK */
  if( led_cfg->gpio_num == 0 )
     return 0;
     
  /* Check if this state has not been already configured */
  if( module->states[ led_cfg->state ] )
    return -EBUSY;

  /* Create a new state */
  if( ( state = kmalloc( sizeof( struct led_state ), GFP_KERNEL ) ) == NULL )
  {
    return -ENOMEM;
  }  
     
  /* Clear memory */
  memset( state, 0, sizeof( struct led_state ) );
  
  /* Initialize LED object */
  if( ( state->object = kmalloc( sizeof( struct led_object ), GFP_KERNEL ) ) == NULL )
  {
    kfree( state );
    
    return -ENOMEM;
  }  
  
  /* Clear memory */
  memset( state->object, 0, sizeof( struct led_object ) );

  state->object->num_of_gpios = led_cfg->gpio_num;
  state->object->led_gpios[ 0 ] = led_cfg->gpio[ 0 ];
  
  /* Initialize LED action */
  if( ( state->action = kmalloc( sizeof( struct led_action ), GFP_KERNEL ) ) == NULL )
  {
    kfree( state->object );
    kfree( state );

    return -ENOMEM;
  }  
         
  /* Initialize LED action fields */
  state->action->color_index = led_cfg->color;
  state->action->param1 = led_cfg->param1;
  state->action->param2 = led_cfg->param2;
  state->action->mode = led_cfg->mode;        

  /* Is this a GPIO LED? */
  if ( state->object->led_gpios[ 0 ] < 128 )
  {
     /* Init multi color GPIO-LED */
     if( state->object->num_of_gpios > 1 )
        state->object->led_gpios[ 1 ] = led_cfg->gpio[ 1 ];
  
     /* Register the GPIO-LED */
     ret = led_hal_register( state->object );
  
     if ( ret < 0 )
     {
        /* Free allocated memory if register failed */
        kfree( state->action );
        kfree( state->object );
        kfree( state );
        
        return ret;
     }
     else
        state->handle = ret;
  }
  else
  {
     state->handle = -1; /* Non-GPIO LED, register later, when callbacks are supplied */
     if ( ( state->object->led_funcs = kmalloc( sizeof( struct led_funcs ), GFP_KERNEL ) ) == NULL )
     {
        /* Free allocated memory if register failed */
        kfree( state->action );
        kfree( state->object );
        kfree( state );
     
        return -ENOMEM;
     }
     
     /* Clear memory */
     memset( state->object->led_funcs, 0, sizeof( struct led_funcs ) );     
  }
       
  /* Save the handle in the action structure for the LED HAL */
  state->action->led_handle = state->handle;

  /* Add a new state to module */
  module->states[ led_cfg->state ] = state;

#ifdef LED_MANAGER_DEBUG
  printk( "%sAdded state %d, got handle: %d\n", LED_MANAGER_MESSAGE_PREFIX, led_cfg->state, state->handle );
#endif       
 
  /* All is OK */
  return 0;
}

/*=============================================================================
   Linux specific implementation
  =============================================================================*/

/************************************************************************
 *  Function: led_manager_ioctl 
 *
 *  Purpose:  IOCTL function for LED parser program
 *                                                                     
 ************************************************************************/
static int led_manager_ioctl( struct inode * inode, struct file * file, 
  unsigned int cmd, unsigned long arg )
{
  int ret = 0;
    
  switch ( cmd )
  {
     case LED_MANAGER_CONFIG:
     {
        struct led_config led_cfg;

        /* Copy configuration structure from user space */
        if (copy_from_user((void *)&led_cfg, (void *)arg, sizeof(struct led_config)))
        {
           ret = -EFAULT;
           break;
        }
        
        ret = led_manager_configure( &led_cfg );
     }
     break;

     case LED_MANAGER_REGISTER:
     {
        struct led_manager_user_module user_module;
        char module_name[ MODULE_NAME_LEN ];
        
        /* Copy module structure from user space */
        if (copy_from_user((void *)&user_module, (void *)arg, sizeof( struct led_manager_user_module ) ))
        {
           ret = -EFAULT;
           break;
        }

        if (copy_from_user((void *)module_name, (void *)user_module.name, MODULE_NAME_LEN ))
        {
           ret = -EFAULT;
           break;
        }
        
        /* Register a module */
        if ( ( user_module.handle = ( unsigned int )led_manager_register_module( module_name, 0 ) ) == 0 )
        {
           ret = -EFAULT;
           break;
        }   
        
        /* Copy the handle back to the user */
        if (copy_to_user(( void *)arg, (void *)&user_module, sizeof( struct led_manager_user_module )))
        {
           ret =  -EFAULT;
           break;
        }
        
#ifdef LED_MANAGER_DEBUG
        printk( "----- returned handle 0x%x\n", user_module.handle );
#endif              
     }
     break;
     
     case LED_MANAGER_UNREGISTER:
     {
        struct led_manager_user_module user_module;
       
        /* Copy module structure from user space */
        if (copy_from_user((void *)&user_module, (void *)arg, sizeof( struct led_manager_user_module ) ))
        {
           ret = -EFAULT;
           break;
        }
        else
        {
#ifdef LED_MANAGER_DEBUG
           printk( "----- unregister handle 0x%x\n", (unsigned int)user_module.handle );
#endif       
           /* Unregister module */
           ret = led_manager_unregister_module( ( void *)user_module.handle );
           
#ifdef LED_MANAGER_DEBUG
           printk( "unregister returned %d\n", ret );
#endif                 
        }
     }
     break;

     case LED_MANAGER_ACTION:
     {
        struct led_manager_user_action user_action;

        /* Copy configuration structure from user space */
        if (copy_from_user((void *)&user_action, (void *)arg, sizeof(struct led_manager_user_action)))
        {
           ret = -EFAULT;
           break;
        }
        else 
        {
#ifdef LED_MANAGER_DEBUG
           printk( "----- received handle 0x%x\n", user_action.handle );    
#endif       
          /* Perform an action */
           ret = led_manager_led_action( ( void *)user_action.handle, user_action.state_id );
           
#ifdef LED_MANAGER_DEBUG
           printk( "action returned %d\n", ret );
#endif       
        }
     }
     break;
     
     
    default:
      ret = -EINVAL;
  }
  return ret;
}


/************************************************************************
 *  Function: led_manager_open 
 *
 *  Purpose:  LED manager open function
 *                                                                     
 ************************************************************************/
static int led_manager_open( struct inode * inode, struct file * file )
{
  return 0;
}

/************************************************************************
 *  Function: led_manager_close 
 *
 *  Purpose:  LED manager close function
 *                                                                     
 ************************************************************************/
static int led_manager_close( struct inode * inode, struct file * file )
{
  return 0;
}

struct file_operations led_manager_fops = {
  ioctl:    led_manager_ioctl,
  open:    led_manager_open,
  release:  led_manager_close
};

devfs_handle_t h;

/************************************************************************
 *  Function: led_manager_init 
 *
 *  Purpose:  LED manager init function
 *                                                                     
 ************************************************************************/
int __init led_manager_init(void)
{
  /* register LED manager device */
  if( ( h = devfs_register(NULL, "led", DEVFS_FL_AUTO_DEVNUM, 0, 0,
      S_IFCHR | S_IRUGO | S_IWUGO, &led_manager_fops,NULL) ) == NULL )
        return -EFAULT;

  /* Initialize modules list head */
  led_module_head = NULL;
  
  return 0;
}

/************************************************************************
 *  Function: led_manager_exit 
 *
 *  Purpose:  LED manager exit function
 *                                                                     
 ************************************************************************/
void __exit led_manager_exit(void)
{
  struct led_module *module = led_module_head;
  struct led_module *cur_module;
  
  /* Unregister LED manager */
  devfs_unregister( h );

  while ( module )
  { 
    cur_module = module;

    /* Save next module in list */
    module = module->next;
    
    /* Unregister module: Free memory and remove from list */
    led_manager_unregister_module( cur_module );
  } 
}

module_init(led_manager_init);
module_exit(led_manager_exit);

EXPORT_SYMBOL_NOVERS(led_manager_led_action);
EXPORT_SYMBOL_NOVERS(led_manager_register_module);
EXPORT_SYMBOL_NOVERS(led_manager_unregister_module);
EXPORT_SYMBOL_NOVERS(led_manager_install_callbacks);
