/*******************************************************************************
**+--------------------------------------------------------------------------+**
**|                            ****                                          |**
**|                            ****                                          |**
**|                            ******o***                                    |**
**|                      ********_///_****                                   |**
**|                      ***** /_//_/ ****                                   |**
**|                       ** ** (__/ ****                                    |**
**|                           *********                                      |**
**|                            ****                                          |**
**|                            ***                                           |**
**|                                                                          |**
**|         Copyright (c) 1998-2005 Texas Instruments Incorporated           |**
**|                        ALL RIGHTS RESERVED                               |**
**|                                                                          |**
**+--------------------------------------------------------------------------+**
*******************************************************************************/

/*******************************************************************************
 * FILE PURPOSE:    Sequencer Serial Port HAL Source
 *******************************************************************************
 * FILE NAME:       ssp_hal.c
 *
 * DESCRIPTION:     HAL Souce code for Sequencer Serial Port
 *
 * REVISION HISTORY:
 * 04 Oct 02 - PSP TII  - Creation
 * 02 Jan 03 - PSP TII  - Removed port 1 support
 * 12 Feb 03 - PSP TII  - Made read and write routines to pause instead of stop
 *                        after each byte is written.
 *
 ******************************************************************************/

#include "haltypes.h"
#include "ssp_hal.h"
#include <pal_os.h>

/* 
 * define the sequencer ram range base
 */
#define SSP_HAL_PT0_SEQ_RAM_LOW                         0x100

/*
 * Define the SSP register masks and 
 * bit positions for
 * IO Select Register.
 */
#define SSP_HAL_IOSEL_PT0_USE_EXTENDED_RAM              0x00008000

/*
 * Define the Port0 interrupt enable
 * masks 
 */
#define SSP_HAL_PORT0_INT_ENABLE                        0x00000001

/* 
 * Define the masks for Port 0 
 * configuration register 2
 */
#define SSP_HAL_PT0_CLK_DIV_MASK                        0x0000000F
#define SSP_HAL_PT0_RAM_WRITE_ENABLE                    0x00000010
#define SSP_HAL_PT0_RAM_READ_ENABLE                     0x00000020
#define SSP_HAL_PT0_16BIT_DATA_REG                      0x00000040
#define SSP_HAL_PT0_AUTO_XFR_START                      0x00000080
#define SSP_HAL_PT0_SOFT_RESET                          0x00000100

/*
 * Define the Port 0/1 signal selection
 * masks.
 */
#define SSP_HAL_DIRECT_INPUT                             0x00
#define SSP_HAL_DIRECT_OUTPUT                            0x04
#define SSP_HAL_PT0_DATA_OUTPUT                          0x01
#define SSP_HAL_PT0_CLK                                  0x02
#define SSP_HAL_PT0_CS                                   0x03

/* 
 * Define the masks for pin 0-5
 * direct output control
 */
#define SSP_HAL_PIN0_DIRECT_OUT_HIGH                     0x00000040
#define SSP_HAL_PIN1_DIRECT_OUT_HIGH                     0x00000080
#define SSP_HAL_PIN2_DIRECT_OUT_HIGH                     0x00000100
#define SSP_HAL_PIN3_DIRECT_OUT_HIGH                     0x00000200
#define SSP_HAL_PIN4_DIRECT_OUT_HIGH                     0x00000400

/*
 * Define the SSP register masks and 
 * bit positions for
 * IO Select Register.
 */
#define SSP_HAL_IOSEL_PIN0_MASK                          0x00000007
#define SSP_HAL_IOSEL_PIN1_MASK                          0x00000038
#define SSP_HAL_IOSEL_PIN2_MASK                          0x000001C0
#define SSP_HAL_IOSEL_PIN3_MASK                          0x00000E00
#define SSP_HAL_IOSEL_PIN4_MASK                          0x00007000
#define SSP_HAL_IOSEL_PT0_USE_EXTENDED_RAM               0x00008000


#define SSP_HAL_IOSEL_PIN0_BIT_POS                       0
#define SSP_HAL_IOSEL_PIN1_BIT_POS                       3
#define SSP_HAL_IOSEL_PIN2_BIT_POS                       6
#define SSP_HAL_IOSEL_PIN3_BIT_POS                       9
#define SSP_HAL_IOSEL_PIN4_BIT_POS                       12

/* 
 * Define the masks for port 0 state
 * register
 */
#define SSP_HAL_PT0_STATE_STOP_ADDR_MASK                0x0000003F         

/*
 * Define the Clock pre divide mask.
 */
#define SSP_HAL_CLK_PRE_DIV_MASK                         0x000000FF

/*
 * define the masks for Port 0 configuration 
 * register 1
 */
#define SSP_HAL_PT0_START_ADDR_MASK                      0x0000003F
#define SSP_HAL_PT0_START_ADDR_VALID_MASK                0x00000080
#define SSP_HAL_PT0_EARLY_DATA_IN_MASK                   0x00000100
#define SSP_HAL_PT0_DELAY_DATA_OUT_MASK                  0x00000200
#define SSP_HAL_PT0_STATE_BUSY_MASK                      0x00000400
#define SSP_HAL_PT0_CLEAR_DATA_REG_MASK                  0x00000800
#define SSP_HAL_PT0_SHIFT_ADDR_LSB_FIRST                 0x00001000
#define SSP_HAL_PT0_SHIFT_DATA_LSB_FIRST                 0x00002000
#define SSP_HAL_PT0_START_SERIAL_TRANSFER                0x00008000

/*
 * Data and address masks
 */
#define SSP_HAL_SPI_SERIAL_DATA_MASK                    0x000000FF
#define SSP_HAL_I2C_SERIAL_DATA_MASK                    0x000000FF
#define SSP_HAL_I2C_READ_ENABLE_BIT                     0x00000100
#define SSP_HAL_I2C_ADDR_MASK                           0x0000FE00
#define SSP_HAL_I2C_ADDR_SHIFT_AMT                      9
  
/*
 * These macros below define some required sequence map 
 * addresses. If the sequence map is changed, the updated 
 * values should be reflected here. 
 */

/* Stop/Pause instructions */
#define SSP_HAL_SEQMAP_I2C_STOP_ADDR                    0x15

/* Read byte instructions */
#define SSP_HAL_SEQMAP_I2C_READ_BYTE_ADDR               0x09
#define SSP_HAL_SEQMAP_I2C_READ_START_ADDR              0x04
#define SSP_HAL_SEQMAP_I2C_READ_ADDR_PAUSE              0x08
#define SSP_HAL_SEQMAP_I2C_READ_BYTE_PAUSE              0x0B

/* Write byte instructions */
#define SSP_HAL_SEQMAP_I2C_WRITE_BYTE_ADDR              0x11
#define SSP_HAL_SEQMAP_I2C_WRITE_START_ADDR             0x0C
#define SSP_HAL_SEQMAP_I2C_WRITE_ADDR_PAUSE             0x10
#define SSP_HAL_SEQMAP_I2C_WRITE_BYTE_PAUSE             0x14


/*
 * The structure representing the set of registers
 * of the SSP device.
 */
typedef struct SSP_HAL_REG_tag
{

    UINT32 revision_control;
    UINT32 iosel_control_1;
    UINT32 iosel_control_2;
    UINT32 clk_pre_divider;
    UINT32 interrupt_status;
    UINT32 interrupt_enable;
    UINT32 test_control;
    UINT32 pad1[9];
    UINT32 port0_config_2;
    UINT32 port0_address;
    UINT32 port0_data;    
    UINT32 port0_config_1;
    UINT32 port0_state;    
    UINT32 pad2[11];
    UINT32 port1_config_2;
    UINT32 port1_address;
    UINT32 port1_data;    
    UINT32 port1_config_1;
    UINT32 port1_state;    
    
}SSP_HAL_REG_T;

/*
 * The SSP HAL object.
 */
typedef struct  SSP_HAL_OBJ_tag
{    
  /* the device base address */
  UINT32 base_addr;

  /* the input frequency */
  UINT32 input_freq;
 
}SSP_HAL_OBJ_T;

/*
 * Module specific global data goes here
 */

/* Create a volatile pointer to help access the SSP registers  */
static volatile SSP_HAL_REG_T *ssp_halp_p_regs;

/*
 * The SSP HAL structure. This stores information
 * on the current device configuration 
 */
static SSP_HAL_OBJ_T ssp_halp_obj;

/* Create pointer to access the SEQ RAM. */
static volatile UINT32 *ssp_halp_p_pt0_ram;

/*
 * The SSP Sequence maps
 */

extern UINT32 ssp_halp_spi_read_seq_map[];

extern UINT32 ssp_halp_spi_write_seq_map[];

extern UINT32 ssp_halp_spi_stop_seq_map[];

extern UINT32 ssp_halp_i2c_seq_map[];

extern UINT32 ssp_halp_lcd_up_seq_map[];

extern UINT32 ssp_halp_lcd_down_seq_map[];

/*******************************************************************************
 * FUNCTION: ssp_halp_load_seq_ram
 *******************************************************************************
 * DESCRIPTION: loads the sequence map into sequence ram
 *
 * RETURNS: returns SSP_HAL_OK on success and SSP_HAL_ERROR on failure.
 ******************************************************************************/
static INT32 ssp_halp_load_seq_ram (UINT32* p_seq_map, 
                                    UINT32 map_size, 
                                    UINT32 start_address);

/*******************************************************************************
 * FUNCTION ssp_hal_init: Initialize the SSP device
 *******************************************************************************
 * DESCRIPTION: This initializes the HAL layer for the SSP module. 
 *              It initializes the SSP registers and the HAL specific 
 *              data structures.
 *
 * RETURNS: returns SSP_HAL_OK on success and SSP_HAL_ERROR on failure. 
 ******************************************************************************/
INT32 ssp_hal_init
(
    UINT32 base_address,
    UINT32 module_input_freq
)
{
    /* Store the information passed */
    ssp_halp_obj.base_addr = base_address;
    ssp_halp_obj.input_freq = module_input_freq;

    /*
     * Initialize the pointers to access the chip's registers 
     */
    ssp_halp_p_regs = (volatile SSP_HAL_REG_T*)base_address;
    ssp_halp_p_pt0_ram = (volatile UINT32*)(base_address + \
                                                SSP_HAL_PT0_SEQ_RAM_LOW);
    
    /*
     * Release the SSP PT0 from soft reset state.
     */
    ssp_halp_p_regs->port0_config_2 &= ~SSP_HAL_PT0_SOFT_RESET;

    /*
     * Enable 16 bit data register and separate busy bit 
     */
    ssp_halp_p_regs->port0_config_2 |= SSP_HAL_PT0_16BIT_DATA_REG;
        
    /*
     * Choose data transfer with kick start.
     */    
    ssp_halp_p_regs->port0_config_2 &= ~SSP_HAL_PT0_AUTO_XFR_START;
            
    /* 
     * Enable MSB First shifting for address and data register 
     */    
    ssp_halp_p_regs->port0_config_1 &=~SSP_HAL_PT0_SHIFT_ADDR_LSB_FIRST;
    ssp_halp_p_regs->port0_config_1 &=~SSP_HAL_PT0_SHIFT_DATA_LSB_FIRST;
    /* Make the entire sequence map accessible from port 0 */
    ssp_halp_p_regs->iosel_control_1|=SSP_HAL_IOSEL_PT0_USE_EXTENDED_RAM;

    /* Make start address valid */    
    ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_START_ADDR_VALID_MASK;

    /* Do not permit reads and writes to sequence RAM */
    ssp_halp_p_regs->port0_config_2 &= ~SSP_HAL_PT0_RAM_WRITE_ENABLE;
    ssp_halp_p_regs->port0_config_2 &= ~SSP_HAL_PT0_RAM_READ_ENABLE;
    
    return SSP_HAL_OK;
}

/*******************************************************************************
 * FUNCTION ssp_hal_i2c_open: Initialize the SSP device for I2C trasactions
 *******************************************************************************
 * DESCRIPTION: Initialize the SSP device for I2C transactions according to
 * the information passed.
 *
 * RETURNS: updated INFO structure in case of success, NULL otherwise.
 ******************************************************************************/
SSP_HAL_I2C_INFO_T* ssp_hal_i2c_open
(
    SSP_HAL_I2C_INFO_T info
)
{
    UINT32 divisor, freq, temp, clk_rate, done;
    SSP_HAL_I2C_INFO_T* ret;
        
    /* Load the I2C sequence map */
    ssp_halp_load_seq_ram(ssp_halp_i2c_seq_map,24,(UINT32)ssp_halp_p_pt0_ram);

    /* make iose1 ctrl 1 and 2 to zero */
    ssp_halp_p_regs->iosel_control_1 = ssp_halp_p_regs->iosel_control_2 = 0;
     
    /* Setup the data pins */
    /* data output: */
    ssp_halp_p_regs->iosel_control_1 |= (0x1 << (3*info.data_pin)) ;
    /* data input: */
    ssp_halp_p_regs->iosel_control_2 |= info.data_pin;

    /* Setup the clock pin */
    ssp_halp_p_regs->iosel_control_1 |= (0x2 << (3*info.clock_pin)) ;

    /* Setup the clock dividers */	
    
	divisor = (UINT32)(ssp_halp_obj.input_freq/info.bus_speed);
    
	if(divisor < 32) {
        ssp_halp_p_regs->port0_config_2 &= ~0xF;
        ssp_halp_p_regs->port0_config_2 |= (divisor & 0xF);
	} else {
        /* peg the divisor at its max value */
		ssp_halp_p_regs->port0_config_2 |= 0xF;	   
	
        /* 
         * now try to further divide using the input clock divisor of
         * the ssp block to see if you can get to your target freq
         */
		freq = ssp_halp_obj.input_freq >> 5;	 /* divide input freq by 256 */
		temp = freq;	 

		clk_rate = 1;
		done = FALSE;
        
		while((temp > info.bus_speed) && (clk_rate < 255) && !done)
		{
			temp = freq / clk_rate;
			if(temp < info.bus_speed) {
				done = TRUE;
            } else {
				clk_rate++;
            }
		}
        
		if(done) {
            ssp_halp_p_regs->clk_pre_divider = clk_rate - 1;
		} else {
            return NULL;
        }
	}
   
    /* reset early data in */
    ssp_halp_p_regs->port0_config_1 &= ~SSP_HAL_PT0_EARLY_DATA_IN_MASK;
    
    /* Set data delay out */
    ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_DELAY_DATA_OUT_MASK; 

    if(info.mode == SSP_HAL_MODE_INTR) {
        ssp_halp_p_regs->interrupt_enable |= SSP_HAL_PORT0_INT_ENABLE;
    } else {
        ssp_halp_p_regs->interrupt_enable &= ~SSP_HAL_PORT0_INT_ENABLE;
    }
        
    if(PAL_osMemAlloc(0, sizeof(SSP_HAL_I2C_INFO_T), 0, ( void *)&ret) != PAL_SOK) {
        return NULL;
    }

    ret->data_pin  = info.data_pin;
    ret->clock_pin = info.clock_pin;
    ret->addr      = info.addr;
    ret->bus_speed = ssp_halp_obj.input_freq/
                        (((ssp_halp_p_regs->port0_config_2 & 0xF) + 1)*2);

    ret->bus_speed = ret->bus_speed/(ssp_halp_p_regs->clk_pre_divider + 1);

    ret->mode      = info.mode;
    
    return ret;
}

/*******************************************************************************
 * FUNCTION ssp_hal_i2c_close: Close the opened I2C device.
 *******************************************************************************
 * DESCRIPTION: Close the opened I2C device.
 *
 * RETURNS: returns SSP_HAL_OK on success and SSP_HAL_ERROR on failure. 
 ******************************************************************************/
INT32 ssp_hal_i2c_close
(
    SSP_HAL_I2C_INFO_T* info
) 
{
    PAL_osMemFree(0, ( void *)info, sizeof(SSP_HAL_I2C_INFO_T));
    return SSP_HAL_OK;
}

/*******************************************************************************
 * FUNCTION ssp_hal_i2c_write: The I2C write method.
 *******************************************************************************
 * DESCRIPTION: The I2C write method. Writes the specified number of bytes and
 *              issues a I2C stop signal. 
 *
 * RETURNS: returns number of bytes written, SSP_HAL_ERROR on error. 
 ******************************************************************************/
INT32 ssp_hal_i2c_write 
(
    SSP_HAL_I2C_INFO_T  *info,
    UINT8               *buffer, 
    UINT32              len
)
{
   
    INT32 count = SSP_HAL_ERROR;

    if(!buffer) {
        return SSP_HAL_ERROR;            
    }
        
    /* Make the sequence map start at the right address */
    ssp_halp_p_regs->port0_config_1 &= (~SSP_HAL_PT0_START_ADDR_MASK);
    ssp_halp_p_regs->port0_config_1 |= (SSP_HAL_SEQMAP_I2C_WRITE_START_ADDR & 
                                            SSP_HAL_PT0_START_ADDR_MASK);

    /* Write the I2C address in the address register */
    ssp_halp_p_regs->port0_address = (info->addr << SSP_HAL_I2C_ADDR_SHIFT_AMT)
                                        & SSP_HAL_I2C_ADDR_MASK;

    /* Enable Write transaction */
    ssp_halp_p_regs->port0_address &= ~SSP_HAL_I2C_READ_ENABLE_BIT;        

    /* Trigger device address write by writing to ST bit */
    ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_START_SERIAL_TRANSFER;
       
    /* Wait for address write to complete */
    if(info->mode == SSP_HAL_MODE_POLL) {        
        for(;ssp_halp_p_regs->port0_config_1 & SSP_HAL_PT0_STATE_BUSY_MASK;);
    } else {
        ssp_drv_cbk_wait_for_xfr_done();
    }
    
    /* check if address write completed successfully*/
    if((ssp_halp_p_regs->port0_state & SSP_HAL_PT0_STATE_STOP_ADDR_MASK) !=
                    SSP_HAL_SEQMAP_I2C_WRITE_ADDR_PAUSE) {
#ifdef SSP_DEBUG            
        printf("Error while writing address for write!"
            " Sequence map stopped at 0x%0x\n", 
            ssp_halp_p_regs->port0_state & SSP_HAL_PT0_STATE_STOP_ADDR_MASK);
#endif
        goto writeerr;
    }
    
    for(count = 0; count < len; count++) {
                    
        /* Make the sequence map start at the right address */
        ssp_halp_p_regs->port0_config_1 &= (~SSP_HAL_PT0_START_ADDR_MASK);
        ssp_halp_p_regs->port0_config_1 |= (SSP_HAL_SEQMAP_I2C_WRITE_BYTE_ADDR                                                 & SSP_HAL_PT0_START_ADDR_MASK);
        
        /* Put the data */
        ssp_halp_p_regs->port0_address = (buffer[count] & 0xFF) <<  8;

        /* Trigger the data write */
        ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_START_SERIAL_TRANSFER;

        /* Wait for data write to complete */
        if(info->mode == SSP_HAL_MODE_POLL) {
            for(;ssp_halp_p_regs->port0_config_1&SSP_HAL_PT0_STATE_BUSY_MASK;);
        } else {
            ssp_drv_cbk_wait_for_xfr_done();
        }
                                                
        /* check if write completed successfully */
        if((ssp_halp_p_regs->port0_state & SSP_HAL_PT0_STATE_STOP_ADDR_MASK)
                        != SSP_HAL_SEQMAP_I2C_WRITE_BYTE_PAUSE) {
            count = SSP_HAL_ERROR;
#ifdef SSP_DEBUG                
            printf("Error while writing byte for write!"
                " Sequence map stopped at 0x%0x\n", 
                ssp_halp_p_regs->port0_state&SSP_HAL_PT0_STATE_STOP_ADDR_MASK);
#endif
               break; 
        }
    }
   
writeerr:    
    /* Signal stop on I2C bus */
    ssp_halp_p_regs->port0_config_1 &= (~SSP_HAL_PT0_START_ADDR_MASK);
    ssp_halp_p_regs->port0_config_1 |= 
                  (SSP_HAL_SEQMAP_I2C_STOP_ADDR & SSP_HAL_PT0_START_ADDR_MASK);
    
    /* Trigger stop signal */
    ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_START_SERIAL_TRANSFER;


    /* Wait for stop to complete */
    if(info->mode == SSP_HAL_MODE_POLL) {
        for(;ssp_halp_p_regs->port0_config_1&SSP_HAL_PT0_STATE_BUSY_MASK;);
    } else {
        ssp_drv_cbk_wait_for_xfr_done();
    }
    return count;
}

/*******************************************************************************
 * FUNCTION ssp_hal_i2c_read: The I2C read method.
 *******************************************************************************
 * DESCRIPTION: The I2C read method. Reads the specified number of bytes and
 *              issues a I2C stop signal. 
 *
 * RETURNS: returns number of bytes written, SSP_HAL_ERROR on error. 
 ******************************************************************************/
INT32 ssp_hal_i2c_read
(
    SSP_HAL_I2C_INFO_T  *info,
    UINT8               *buffer,
    UINT32              len
)
{
    UINT32 count = SSP_HAL_ERROR;

    if(!buffer) {
        return SSP_HAL_ERROR;            
    }

    /* Make the sequence map start at the right address */
    ssp_halp_p_regs->port0_config_1 &= (~SSP_HAL_PT0_START_ADDR_MASK);
    ssp_halp_p_regs->port0_config_1 |= (SSP_HAL_SEQMAP_I2C_READ_START_ADDR 
                                            & SSP_HAL_PT0_START_ADDR_MASK);
    
    /* Write the I2C address in the address register */
    ssp_halp_p_regs->port0_address = (info->addr << SSP_HAL_I2C_ADDR_SHIFT_AMT) 
                                            & SSP_HAL_I2C_ADDR_MASK;    

    /* Enable Read transaction */
    ssp_halp_p_regs->port0_address |= SSP_HAL_I2C_READ_ENABLE_BIT;        
   
    /* Trigger device address write */
    ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_START_SERIAL_TRANSFER;

    /* Wait for address write to complete */
    if(info->mode == SSP_HAL_MODE_POLL) {
        for(;ssp_halp_p_regs->port0_config_1&SSP_HAL_PT0_STATE_BUSY_MASK;);
    } else {
        ssp_drv_cbk_wait_for_xfr_done();
    }
    
    if((ssp_halp_p_regs->port0_state & SSP_HAL_PT0_STATE_STOP_ADDR_MASK) != 
                    SSP_HAL_SEQMAP_I2C_READ_ADDR_PAUSE)
    {
#ifdef SSP_DEBUG            
        printf("Error while writing address for read!" 
            " Sequence map stopped at 0x%0x\n", 
            ssp_halp_p_regs->port0_state & SSP_HAL_PT0_STATE_STOP_ADDR_MASK);
#endif
        goto readerr;              
    }
    
    for(count = 0; count < len; count++) {
            
        /* Clear the data register */
        ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_CLEAR_DATA_REG_MASK;            
            
        /* Make the sequence map start at the right address */
        ssp_halp_p_regs->port0_config_1 &= (~SSP_HAL_PT0_START_ADDR_MASK);
        ssp_halp_p_regs->port0_config_1 |= (SSP_HAL_SEQMAP_I2C_READ_BYTE_ADDR 
                                                & SSP_HAL_PT0_START_ADDR_MASK);
            
        /* Trigger the data read */        
        ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_START_SERIAL_TRANSFER;

        /* Wait for data read to complete */
        if(info->mode == SSP_HAL_MODE_POLL) {
            for(;ssp_halp_p_regs->port0_config_1&SSP_HAL_PT0_STATE_BUSY_MASK;);
        } else {
            ssp_drv_cbk_wait_for_xfr_done();
        }
                
        /* check if read completed successfully */
        if((ssp_halp_p_regs->port0_state & 
                SSP_HAL_PT0_STATE_STOP_ADDR_MASK) != 
                        SSP_HAL_SEQMAP_I2C_READ_BYTE_PAUSE) {
            count = SSP_HAL_ERROR;
#ifdef SSP_DEBUG
        printf("Error while reading byte! Sequence map stopped at 0x%2x\n",
            ssp_halp_p_regs->port0_state & SSP_HAL_PT0_STATE_STOP_ADDR_MASK);
#endif
            break;                          
        }
        
        /* collect the data */
        buffer[count] = ssp_halp_p_regs->port0_data &= 
                                SSP_HAL_I2C_SERIAL_DATA_MASK;        
    }

readerr:   
    /* Signal stop on I2C bus */
    ssp_halp_p_regs->port0_config_1 &= (~SSP_HAL_PT0_START_ADDR_MASK);
    ssp_halp_p_regs->port0_config_1 |= (SSP_HAL_SEQMAP_I2C_STOP_ADDR 
                                                & SSP_HAL_PT0_START_ADDR_MASK);
    
    /* Trigger stop signal */
    ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_START_SERIAL_TRANSFER;

    /* Wait for stop to complete */
    if(info->mode == SSP_HAL_MODE_POLL) {
        for(;ssp_halp_p_regs->port0_config_1&SSP_HAL_PT0_STATE_BUSY_MASK;);
    } else {
        ssp_drv_cbk_wait_for_xfr_done();
    }
    
    return count;
}

/*******************************************************************************
 * FUNCTION ssp_hal_spi_open: Initialize the SSP device for SPI trasactions
 *******************************************************************************
 * DESCRIPTION: Initialize the SSP device for SPI transactions according to
 * the information passed.
 *
 * RETURNS: updated INFO structure in case of success, NULL otherwise.
 ******************************************************************************/
SSP_HAL_SPI_INFO_T* ssp_hal_spi_open
(
    SSP_HAL_SPI_INFO_T info
)
{
    UINT32 pre_div;
    
    SSP_HAL_SPI_INFO_T* ret;
        
    /* re-init iose1 ctrl 1 and 2*/
    
    /* Make the entire sequence map accessible from port 0 */
    ssp_halp_p_regs->iosel_control_1 = SSP_HAL_IOSEL_PT0_USE_EXTENDED_RAM;
    ssp_halp_p_regs->iosel_control_2 = 0;
     
    /* Setup the data pins */
    /* data output: */
    ssp_halp_p_regs->iosel_control_1 |= (0x1 << (3*info.data_out_pin)) ;
    /* data input: */
    ssp_halp_p_regs->iosel_control_2 |= info.data_in_pin;

    /* Setup the clock pin */
    ssp_halp_p_regs->iosel_control_1 |= (0x2 << (3*info.clock_pin)) ;

    /* Setup the chip select pin */
    ssp_halp_p_regs->iosel_control_1 |= (0x3 << (3*info.cs_pin)) ;
    
    /* Setup the clock dividers */	

    /* cuts pre divided frequency by one half */
    ssp_halp_p_regs->port0_config_2 &= ~0xF;
        
    /* requested frequency is doubled to account for the division above */
    pre_div = ssp_halp_obj.input_freq/(info.clock_speed * 2) + 
                    ((ssp_halp_obj.input_freq % (info.clock_speed * 2)) != 0);
    
    ssp_halp_p_regs->clk_pre_divider = (pre_div == 0)? 0: (pre_div - 1);  
   
    /* reset data delay out */
    ssp_halp_p_regs->port0_config_1 &= ~SSP_HAL_PT0_DELAY_DATA_OUT_MASK; 

    /* set early data in */
    ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_EARLY_DATA_IN_MASK;

    if(info.mode == SSP_HAL_MODE_INTR) {
        ssp_halp_p_regs->interrupt_enable |= SSP_HAL_PORT0_INT_ENABLE;
    } else {
        ssp_halp_p_regs->interrupt_enable &= ~SSP_HAL_PORT0_INT_ENABLE;
    }
        
    if(PAL_osMemAlloc(0, sizeof(SSP_HAL_SPI_INFO_T), 0, ( void *)&ret) != PAL_SOK) {
        return NULL;    
    }

    ret->data_in_pin  = info.data_in_pin;
    ret->data_out_pin = info.data_out_pin;
    ret->clock_pin    = info.clock_pin;
    ret->cs_pin       = info.cs_pin;
    ret->clock_speed  = ssp_halp_obj.input_freq/
                            (((ssp_halp_p_regs->port0_config_2 & 0xF) + 1)*2);
    ret->clock_speed  = ret->clock_speed/(ssp_halp_p_regs->clk_pre_divider + 1);
    ret->mode         = info.mode;
    
#ifdef SSP_DEBUG    
    printf("Data in pin: %d\n", ret->data_in_pin);
    printf("Data out pin: %d\n", ret->data_out_pin);
    printf("Clock pin: %d\n", ret->clock_pin);
    printf("CS pin: %d\n", ret->cs_pin);
    printf("Speed: %d\n", ret->clock_speed);
    printf("Mode: %d\n", ret->mode);
#endif
    
    return ret;
}

/*******************************************************************************
 * FUNCTION ssp_hal_spi_close: Close the opened SPI device.
 *******************************************************************************
 * DESCRIPTION: Close the opened SPI device.
 *
 * RETURNS: returns SSP_HAL_OK on success and SSP_HAL_ERROR on failure. 
 ******************************************************************************/
INT32 ssp_hal_spi_close
(
    SSP_HAL_SPI_INFO_T* info
) 
{
    PAL_osMemFree(0, ( void *)info, sizeof(SSP_HAL_SPI_INFO_T));
    return SSP_HAL_OK;
}

/*******************************************************************************
 * FUNCTION ssp_hal_spi_write_read: The SPI write and read method.
 *******************************************************************************
 * DESCRIPTION: The SPI write and method. Writes the specified number of bytes
 *              and then reads the specified number of bytes.
 *
 * RETURNS: returns number of bytes read and written, SSP_HAL_ERROR on error. 
 ******************************************************************************/
INT32 ssp_hal_spi_write_read 
(
    SSP_HAL_SPI_INFO_T  *info,
    UINT8               *write_buf, 
    UINT32              write_len,
    UINT8               *read_buf, 
    UINT32              read_len    
)
{
   
    UINT32 count = 0, ret;

    if((!read_buf) && (!write_buf)) {
        return SSP_HAL_ERROR;            
    }
    
    /* Make the sequence map start at address 0 */
    ssp_halp_p_regs->port0_config_1 &= (~SSP_HAL_PT0_START_ADDR_MASK);
    
    if((write_buf) && (write_len > 0)) {
        ssp_halp_load_seq_ram(ssp_halp_spi_write_seq_map, 2, 
                                        (UINT32)ssp_halp_p_pt0_ram);
        
        for(count = 0; count < write_len; count++) {
            
            /* Put the data */
            ssp_halp_p_regs->port0_address = (write_buf[count] & 0xFF) <<  8;

            /* Trigger the data write */
            ssp_halp_p_regs->port0_config_1|=SSP_HAL_PT0_START_SERIAL_TRANSFER;

            /* Wait for data write to complete */
            if(info->mode == SSP_HAL_MODE_POLL) {
                for(;ssp_halp_p_regs->port0_config_1 & 
                                        SSP_HAL_PT0_STATE_BUSY_MASK;);
            } else {
                ssp_drv_cbk_wait_for_xfr_done();
            }        
        }                    
    } 
    
    ret = count;
    
    if((read_buf) && (read_len > 0)) {
        ssp_halp_load_seq_ram(ssp_halp_spi_read_seq_map, 2, 
                                       (UINT32)ssp_halp_p_pt0_ram);
        for(count = 0; count < read_len; count++) {
                
            /* clear the address register */
            ssp_halp_p_regs->port0_address = 0;
            
            /* Trigger the data read */
            ssp_halp_p_regs->port0_config_1|=SSP_HAL_PT0_START_SERIAL_TRANSFER;

            /* Wait for data read to complete */
            if(info->mode == SSP_HAL_MODE_POLL) {
                for(;ssp_halp_p_regs->port0_config_1 & 
                                        SSP_HAL_PT0_STATE_BUSY_MASK;);
            } else {
                ssp_drv_cbk_wait_for_xfr_done();
            }        
            
            /* Collect the data */
            read_buf[count] = ssp_halp_p_regs->port0_address &= 
                                            SSP_HAL_SPI_SERIAL_DATA_MASK;
                    
        }
    }        

    ret += count;
            
    /* jump to SPI Stop condition */
    ssp_halp_load_seq_ram(ssp_halp_spi_stop_seq_map, 1, 
                                        (UINT32)ssp_halp_p_pt0_ram);
    
    /* Trigger stop signal */
    ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_START_SERIAL_TRANSFER;

    /* Wait for stop to complete */
    if(info->mode == SSP_HAL_MODE_POLL) {
        for(;ssp_halp_p_regs->port0_config_1&SSP_HAL_PT0_STATE_BUSY_MASK;);
    } else {
        ssp_drv_cbk_wait_for_xfr_done();
    }
    
    return ret;
}


/*******************************************************************************
 * FUNCTION NAME : ssp_hal_lcd_ctrl_open: SSP HAL open function
 *******************************************************************************
 * DESCRIPTION   : The SSP HAL LCD contrast control pulse function.
 *
 * RETURNS: updated INFO structure in case of success, NULL otherwise.
 *
 ******************************************************************************/
SSP_HAL_LCD_CTRL_INFO_T* ssp_hal_lcd_ctrl_open
(
    SSP_HAL_LCD_CTRL_INFO_T    info
)
{
    SSP_HAL_LCD_CTRL_INFO_T *ret;
    
    if(PAL_osMemAlloc(0, sizeof(SSP_HAL_LCD_CTRL_INFO_T), 0, ( void *)&ret) != PAL_SOK) {
        return NULL;
    }
    
    PAL_osMemCopy( ret, &info, sizeof(SSP_HAL_LCD_CTRL_INFO_T) );
    
    return ret;
}

/*******************************************************************************
 * FUNCTION NAME : ssp_hal_lcd_ctrl_send: SSP HAL control pulse function
 *******************************************************************************
 * DESCRIPTION   : The SSP HAL LCD contrast control pulse function.
 *
 * RETURNS:      SSP_HAL_OK on success, SSP_HAL_ERROR on  
 *
 ******************************************************************************/
INT32 ssp_hal_lcd_ctrl_send
(
    SSP_HAL_LCD_CTRL_INFO_T    *info,
    BOOL up
)
{
    UINT32*  p_seq_map;
    UINT32             seq_map_size;
    UINT32 pre_div;
    UINT32 clk_div;
            
    /* point to correct (up or down) sequence map */
    p_seq_map = (up == TRUE) ? ssp_halp_lcd_up_seq_map : ssp_halp_lcd_down_seq_map;

    /* both up and down maps same size */
    seq_map_size = 7;

    /* Load the sequence ram */
    ssp_halp_load_seq_ram(p_seq_map, seq_map_size, (UINT32)ssp_halp_p_pt0_ram);
                                                    
    /* make start address valid */  
    ssp_halp_p_regs->port0_config_1 |= SSP_HAL_PT0_START_ADDR_VALID_MASK;
    for(;ssp_halp_p_regs->port0_config_1 & SSP_HAL_PT0_STATE_BUSY_MASK;); 
        
    /* get port 0 out of the soft reset state if it is in it. */
    ssp_halp_p_regs->port0_config_2 &= ~SSP_HAL_PT0_SOFT_RESET;
    
    /* use only the pre divide to get the resolution */
    clk_div = 0; /* cuts pre divided frequency by one half */
    
    /* requested frequency is doubled to account for the division above */
    pre_div = (ssp_halp_obj.input_freq/(info->clock_speed*2) + \
                ((ssp_halp_obj.input_freq % (info->clock_speed*2)) != 0));
    
    pre_div -= 1;    
                                                
    /* reset the delay dout, if set previously */
    ssp_halp_p_regs->port0_config_1 &= ~SSP_HAL_PT0_DELAY_DATA_OUT_MASK;

    /* reset the shift data in early, if set previously*/ 
    ssp_halp_p_regs->port0_config_1 &= ~SSP_HAL_PT0_EARLY_DATA_IN_MASK;
                                                
    /* set the clock div and the pre div */
    pre_div &= SSP_HAL_CLK_PRE_DIV_MASK;
    ssp_halp_p_regs->clk_pre_divider &= ~SSP_HAL_CLK_PRE_DIV_MASK;
    ssp_halp_p_regs->clk_pre_divider |= pre_div;

    clk_div &= SSP_HAL_PT0_CLK_DIV_MASK;
    ssp_halp_p_regs->port0_config_2 &= ~SSP_HAL_PT0_CLK_DIV_MASK;
    ssp_halp_p_regs->port0_config_2 |= clk_div;
                                                
    /* Setup the data pins */
    /* data output: */
    ssp_halp_p_regs->iosel_control_1 |= (SSP_HAL_PT0_DATA_OUTPUT << (3*info->data_out_pin)) ;
    /* data input: */
    ssp_halp_p_regs->iosel_control_2 |= info->data_in_pin;

    /* Setup the clock pin */
    ssp_halp_p_regs->iosel_control_1 |= (SSP_HAL_PT0_CLK << (3*info->clock_pin)) ;

    /* Setup the chip select pin */
    ssp_halp_p_regs->iosel_control_1 |= (SSP_HAL_PT0_CS << (3*info->cs_pin)) ;
    
    /* If the SSP was configured to work with interrupts, disable them */
    if(info->mode == SSP_HAL_MODE_INTR)
    {
        /*ssp_halp_set_param_ctl(SSP_HAL_SET_PT0_INT_DISABLE, 0);*/
        ssp_halp_p_regs->interrupt_enable &= ~SSP_HAL_PORT0_INT_ENABLE;
        
    }
    
    /* Set port 0 TX start on ST bit. */
    ssp_halp_p_regs->port0_config_2 &= \
                        ~SSP_HAL_PT0_AUTO_XFR_START;
    
    /* Start port 0 transfer. */
    ssp_halp_p_regs->port0_config_1 |=SSP_HAL_PT0_START_SERIAL_TRANSFER;
    
    /* Wait for transfer to finish */
    for(;ssp_halp_p_regs->port0_config_1 & SSP_HAL_PT0_STATE_BUSY_MASK;);    

    ssp_halp_p_regs->iosel_control_1 &= ~SSP_HAL_IOSEL_PIN4_MASK;
    ssp_halp_p_regs->iosel_control_1 |= (SSP_HAL_DIRECT_OUTPUT << \
                                            SSP_HAL_IOSEL_PIN4_BIT_POS);

    ssp_halp_p_regs->iosel_control_2 |= SSP_HAL_PIN4_DIRECT_OUT_HIGH;
     
    /* If interrupts were disabled, re-enble them. */
    if(info->mode == SSP_HAL_MODE_INTR)
    {        
        ssp_halp_p_regs->interrupt_enable |= SSP_HAL_PORT0_INT_ENABLE;
    }

    return SSP_HAL_OK;    
}

/*******************************************************************************
 * FUNCTION ssp_hal_lcd_ctrl_close: Close the opened LCD device.
 *******************************************************************************
 * DESCRIPTION: Close the opened LCD device.
 *
 * RETURNS: returns SSP_HAL_OK on success and SSP_HAL_ERROR on failure. 
 ******************************************************************************/
INT32 ssp_hal_lcd_ctrl_close
(
    SSP_HAL_LCD_CTRL_INFO_T* info
) 
{
    PAL_osMemFree(0, ( void *)info, sizeof(SSP_HAL_LCD_CTRL_INFO_T));
    return SSP_HAL_OK;
}

/*******************************************************************************
 * FUNCTION:    ssp_halp_load_seq_ram
 *******************************************************************************
 * DESCRIPTION: loads the sequence map into sequence ram
 *
 * RETURNS: returns SSP_HAL_OK on success and SSP_HAL_ERROR on failure.
 ******************************************************************************/
static INT32 ssp_halp_load_seq_ram
(
    UINT32      *p_seq_map,
    UINT32      map_size,
    UINT32      start_address
)
{
    volatile UINT32 *p_ssp_ram = (volatile UINT32*)start_address; 
    UINT32  i ; 

    /* set the sequential ram of port 0 to be writable */
    ssp_halp_p_regs->port0_config_2 |= SSP_HAL_PT0_RAM_WRITE_ENABLE;
    
    /* put in new seq map */
    for(i = 0 ; i < map_size; i++)
    {
        p_ssp_ram[i]  = p_seq_map[i];
    }

    /* disable RAM reads/writes again */
    ssp_halp_p_regs->port0_config_2 &= ~SSP_HAL_PT0_RAM_WRITE_ENABLE;

    return SSP_HAL_OK;
}

/*******************************************************************************
 * FUNCTION:    ssp_hal_intr_ctrl
 *******************************************************************************
 * DESCRIPTION: Device level interrupt control for SSP module.
 *
 * RETURNS: returns SSP_HAL_OK on success and SSP_HAL_ERROR on failure.
 ******************************************************************************/
INT32 ssp_hal_intr_ctrl
(
    UINT32 action
)
{
     switch (action) {
        case SSP_HAL_INTR_ACK:
            ssp_halp_p_regs->interrupt_status |= 0x1;
        break;
        case SSP_HAL_INTR_DISABLE:
            ssp_halp_p_regs->interrupt_enable &= ~SSP_HAL_PORT0_INT_ENABLE;
        break;
        case SSP_HAL_INTR_ENABLE:
            ssp_halp_p_regs->interrupt_enable |= SSP_HAL_PORT0_INT_ENABLE;
        break;
        default:
            return SSP_HAL_ERROR;
     }

     return SSP_HAL_OK;
}


#ifdef SSP_DEBUG

/*******************************************************************************
 * FUNCTION: ssp_halp_print_regs
 *******************************************************************************
 *
 * DESCRIPTION: Print all of SSP registers
 *
 * RETURNS:     N/A
 ******************************************************************************/
static void ssp_halp_print_regs()
{
    printf("\nIOSEL REG 1: 0x%08x\n",ssp_halp_p_regs->iosel_control_1);
    printf("IOSEL REG 2: 0x%08x\n",ssp_halp_p_regs->iosel_control_2);
    printf("CLK PRE DIV: 0x%08x\n",ssp_halp_p_regs->clk_pre_divider);
    printf("INT STAT   : 0x%08x\n",ssp_halp_p_regs->interrupt_status);
    printf("INT ENABLE : 0x%08x\n",ssp_halp_p_regs->interrupt_enable);
    printf("CONFIG 1   : 0x%08x\n",ssp_halp_p_regs->port0_config_1);
    printf("CONFIG 2   : 0x%08x\n",ssp_halp_p_regs->port0_config_2);
    printf("DATA       : 0x%08x\n",ssp_halp_p_regs->port0_data);
    printf("ADDR       : 0x%08x\n",ssp_halp_p_regs->port0_address);
    printf("STATE      : 0x%08x\n\n",ssp_halp_p_regs->port0_state);        
}

#endif
