//-----------------------------------------------------------------------------------------------------
// Copyright 2003, Texas Instruments Incorporated
//
// This program has been modified from its original operation by Texas Instruments
// to do the following:
//       To drive the GLCD display.
//
//   THIS MODIFIED SOFTWARE AND DOCUMENTATION ARE PROVIDED
//   "AS IS," AND TEXAS INSTRUMENTS MAKES NO REPRESENTATIONS 
//   OR WARRENTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
//   TO, WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY 
//   PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE OR 
//   DOCUMENTATION WILL NOT INFRINGE ANY THIRD PARTY PATENTS, 
//   COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS. 
//   See The GNU General Public License for more details.
//
// These changes are covered under version 2 of the GNU General Public License, 
//  dated June 1991.
//-----------------------------------------------------------------------------------------------------

/*
 * File: fb.c
 *
 * The frame buffer driver for TI MIPS boards: tnetv1050.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS  PROVIDED  ``AS  IS''  AND   ANY  EXPRESS  OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Please report all bugs and problems to the author or <support@dsplinux.net>
 *
 * key: RRGPLCR (do not remove)
 *
 */

#include "glcd.h"
#include <linux/init.h>
#include <linux/slab.h>
#include <video/fbcon.h>
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb2.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <pal.h>
#include "glcd_hal.h"
#include "glcd_hal_apis.h"

#ifdef CONFIG_GLCD_OPTREX_MONO
#include "optrex_mono.h"
#elif CONFIG_GLCD_SHARP_COLOR
#include "sharp_color.h"
#ifdef FB_DEBUG
#include "glcd_test.h"
#endif
#else
#include "optrex_mono.h"
#endif

#define CACHELN_SZ 16

#define PALETTE_SIZE 256		// Size of the Palette
#define LEFT_MARGIN 64		// Left Margin
#define RIGHT_MARGIN 64		// Right Margin
#define UPPER_MARGIN 32		// Upper Margin
#define LOWER_MARGIN 32		// Lower Margin
#define NUM_XPIXELS 320
#define NUM_YPIXELS 240

/* Device Dependent Information of the Particular Hardware */
struct tifb_par
{
        unsigned char *p_screen_base;
        unsigned char *v_screen_base;
        unsigned char *p_palette_base;
        unsigned char *v_palette_base;
        unsigned long screen_size;
        unsigned int  palette_size;
        unsigned int  currcon;
        unsigned int  bpp;
};

static u_long tifb_memory;
static int buf_size;
static struct display disp;
static struct fb_info tifb_info;
struct tifb_par current_par;

static void set_color_bitfields(struct fb_var_screeninfo *var);

/* Check that we have frame buffer support */
#if defined( FBCON_HAS_MFB )
#define FBCON_OK
#elif defined( FBCON_HAS_CFB2 )
#define FBCON_OK
#elif defined( FBCON_HAS_CFB4 )
#define FBCON_OK
#elif defined( FBCON_HAS_CFB8 )
#define FBCON_OK
#elif defined( FBCON_HAS_CFB16 )
#define FBCON_OK
#endif

#ifndef FBCON_OK
#error No frame buffer support selected!
#endif

/* Number of Palette Entries */
static union {
#ifdef FBCON_HAS_MFB
    u16 mfb[16];
#endif
#ifdef FBCON_HAS_CFB2
    u16 cfb2[16];
#endif
#ifdef FBCON_HAS_CFB4
    u16 cfb4[16];
#endif
#ifdef FBCON_HAS_CFB8
    u16 cfb8[16];
#endif
#ifdef FBCON_HAS_CFB16
    u16 cfb16[16];
#endif
} fbcon_cmap;

static const char desc[]   = "TI MIPS Framebuffer Driver";
static int tifb_enable = 1;

/* Frame Buffer Operations */
static struct fb_ops tifb_ops =
{
        owner:          THIS_MODULE,
        fb_get_fix:     tifb_get_fix,
        fb_get_var:     tifb_get_var,
        fb_set_var:     tifb_set_var,
        fb_get_cmap:    tifb_get_cmap,
        fb_set_cmap:    tifb_set_cmap,
        fb_pan_display: tifb_pan_display,
};


/* To check for valid displays in set_var */
struct _displays
{
        int xres;
        int yres;
        int bpp;
};


/* Supported Displays by the Driver */
static struct _displays displays_supported[] = 
{ 
#ifdef FBCON_HAS_MFB
  { NUM_XPIXELS, NUM_YPIXELS, 1 },
#endif
#ifdef FBCON_HAS_CFB2
  { NUM_XPIXELS, NUM_YPIXELS, 2 },
#endif
#ifdef FBCON_HAS_CFB4
  { NUM_XPIXELS, NUM_YPIXELS, 4 },
#endif
#ifdef FBCON_HAS_CFB8
  { NUM_XPIXELS, NUM_YPIXELS, 8 },
#endif
#ifdef FBCON_HAS_CFB16
  { NUM_XPIXELS, NUM_YPIXELS, 16 }, 
#endif
};


/* Variable Screen Information */
static struct fb_var_screeninfo tifb_default =
{
        xoffset: 0,
        yoffset: 0,
        transp: {0, 0, 0},
        nonstd: 0,
        activate: 0,
        height: -1,
        width: -1,
        accel_flags: 0,
        left_margin: LEFT_MARGIN,
        right_margin: RIGHT_MARGIN,
        upper_margin: UPPER_MARGIN,
        lower_margin: LOWER_MARGIN,
        sync: 0,
        vmode: FB_VMODE_NONINTERLACED
};


/* Get the Fixed Part of the Display */
static int tifb_get_fix(struct fb_fix_screeninfo *fix, int con, 
                        struct fb_info *info)
{
        struct fb_var_screeninfo *var;

        if ( con == -1 ){
		           var = &tifb_default;
        }
        else
                var = &fb_display[con].var;

        tifb_encode_fix(fix, var);
        return 0;
}


/* Get the User Defined Part of the Display */
static int tifb_get_var(struct fb_var_screeninfo *var, int con, 
                        struct fb_info *info)
{
    if (con == -1)
	      *var = tifb_default;
    else
	      *var = fb_display[con].var;
    
    set_color_bitfields(var);
    return 0;
}


/* Set the User Defined Part of the Display */
static int tifb_set_var(struct fb_var_screeninfo *var, int con, 
                        struct fb_info *info)
{
        struct display *display;
        int i;
        int ok_display;

        // Make sure we support the new Display
        ok_display = 0;
        for ( i=0; i < (sizeof(displays_supported)/sizeof(struct _displays)); i++ ){
                if ( (var->xres == displays_supported[i].xres) &&
                     (var->yres == displays_supported[i].yres) &&
                     (var->bits_per_pixel == displays_supported[i].bpp) ){
                        ok_display = 1;
                        break;
                }
        }

        if ( ! ok_display )
                return -EINVAL;
        
	if ( con >= 0 )
		display = &fb_display[con];
	else
		display = info->disp;	// Used during Initialization

	// Update the Display Interface
	display->var            = *var;
  var->bits_per_pixel     = lcd_cfg.bpp; /* Can't change BPP */
	display->screen_base	= current_par.v_screen_base;

  switch (var->bits_per_pixel) {
	 case 1:
	     display->visual = FB_VISUAL_MONO01;
	     break;
	 case 2:
	 case 4:
	 case 8:
	     display->visual = FB_VISUAL_PSEUDOCOLOR;
	     break;
	 case 16:
	     display->visual = FB_VISUAL_TRUECOLOR;
	     break;
   default:
       return -EINVAL; 
       
	   }


	display->type		= FB_TYPE_PACKED_PIXELS;
	display->type_aux	= 0;
	display->ypanstep	= 0;
	display->ywrapstep	= 0;
	display->line_length	= (var->xres_virtual * var->bits_per_pixel) / 8; 
	display->next_line      = (var->xres_virtual * var->bits_per_pixel) / 8;
	display->can_soft_blank = 1;
	display->inverse	= 0;

#ifdef FB_DEBUG
	printk(KERN_INFO "GLCD: display->bpp = %d\n", display->var.bits_per_pixel);
#endif
	
	switch (var->bits_per_pixel) {
#ifdef FBCON_HAS_MFB
	case 1:
	   display->dispsw = &fbcon_mfb;
	   break;
#endif
#ifdef FBCON_HAS_CFB2
	case 2:
	   display->dispsw = &fbcon_cfb2;
	   break;
#endif
#ifdef FBCON_HAS_CFB4
	case 4:
	   display->dispsw = &fbcon_cfb4;
	   break;
#endif
#ifdef FBCON_HAS_CFB8
	case 8:
	   display->dispsw = &fbcon_cfb8;
	   break;
#endif
#ifdef FBCON_HAS_CFB16
	case 16:
	   display->dispsw = &fbcon_cfb16;
	   display->dispsw_data = fbcon_cmap.cfb16;
	   break;
#endif
	default:
#ifdef FB_DEBUG
		printk(KERN_ALERT "GLCD: Color Depth not Supported!\n");
#endif
		display->dispsw = &fbcon_dummy;
		break;
	}
  
	return 0;
}


/* Get the Display's Colormap */
static int tifb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
                         struct fb_info *info)
{
	// Is Current Console?
	if ( (!info->display_fg) || (con == info->display_fg->vc_num) ){ 
		return fb_get_cmap(cmap, kspc, tifb_getcolreg, info);
        }
	// Non Default Colormap?
	else if ( fb_display[con].cmap.len ){
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
        }
	else{
	fb_copy_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel),
		     cmap, kspc ? 0 : 2);
	}

	return 0;
}


/* Set the Display's Colormap */
static int tifb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
                         struct fb_info *info)
{
        int err;
	struct display *disp;

	if ( con >= 0 )
		disp = &fb_display[con];
	else
		disp = info->disp;

	// No Colormap Allocated?
	if ( !disp->cmap.len ){ 
		// Allocate the Colormap and Return if Error
  	if ((err = fb_alloc_cmap(&disp->cmap,
			      1<<fb_display[con].var.bits_per_pixel, 0)))
	    return err;
	}

	if ( !info->display_fg || con == info->display_fg->vc_num )
		return fb_set_cmap(cmap, kspc, tifb_setcolreg, info);
	else
		fb_copy_cmap(cmap, &disp->cmap, kspc ? 0 : 1);

	/* Flush the buffer to the screen. */
	lcd_hal_blit(LOAD_DATA, 
				( char *)current_par.p_screen_base );

	return 0;
}


/* Pan or Wrap the Display */
static int tifb_pan_display(struct fb_var_screeninfo *var, int con,
                            struct fb_info *info)
{
	// Do Nothing
        return 0;
}


/* Enocde the Unchangable Information */
static void tifb_encode_fix(struct fb_fix_screeninfo *fix,
                            struct fb_var_screeninfo *var)
{
        memset(fix, 0, sizeof(struct fb_fix_screeninfo));
        strcpy(fix->id, desc);
        fix->smem_start  = (unsigned long) current_par.p_screen_base;
        fix->smem_len    = ((lcd_cfg.p_disp_panel->width * lcd_cfg.p_disp_panel->height * lcd_cfg.bpp) / 8);
        fix->type        = FB_TYPE_PACKED_PIXELS;
        fix->type_aux    = 0;

        switch (var->bits_per_pixel) {
      case 1:
          fix->visual = FB_VISUAL_MONO01;
          break;
      case 2:
      case 4:
      case 8:
          fix->visual = FB_VISUAL_PSEUDOCOLOR;
          break;
      case 16:
          fix->visual = FB_VISUAL_TRUECOLOR;
          break;
        }

        fix->ywrapstep   = 1;
        fix->xpanstep    = 1;
        fix->ypanstep    = 1;
        fix->line_length = (var->xres_virtual * var->bits_per_pixel) / 8;
        fix->accel       = FB_ACCEL_NONE;
}


/* Setup the Frame Buffer */
int __init tifb_setup(char *options) 
{
	// Do Nothing
	return 0;
}


/* Console Switching */
static int tifbcon_switch(int con, struct fb_info *info)
{
	// Do Nothing
	return 0;
}


/* Blank the Display */
static void tifbcon_blank(int blank, struct fb_info *info)
{
	// Turn Off the Display
        if (blank) {
		lcd_disable();
                hardware_disable();
        }

	// Turn back On
        else {
                hardware_enable();
                lcd_enable();
        }
}


/* Read a Single Color Register and Split it into Colors/Transparent */
static int tifb_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                          u_int *transp, struct fb_info *info)
{
	u_short pal_entry;
	unsigned short *palette = (unsigned short *)current_par.v_palette_base;

	if ( regno > (PALETTE_SIZE - 1) )
		return 1;

	pal_entry = palette[regno];

	*red    = (pal_entry & 0x000f) << 12;
	*green  = (pal_entry & 0x00f0) << 8;
	*blue   = (pal_entry & 0x0f00) << 4;
	*transp = 0;

	return 0;
}


/* Set a Single Color Register */
static int tifb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                          u_int transp, struct fb_info *info)
{
	u_short pal;
	unsigned short *palette = (unsigned short *)current_par.v_palette_base;

	if ( regno > (PALETTE_SIZE - 1) )
		return 1;

	pal   = ((red   >>  4) & 0xf00);
	pal  |= ((green >>  8) & 0x0f0);
	pal  |= ((blue  >> 12) & 0x00f);
        
	palette[regno] = pal;

        switch ( current_par.bpp ) 
        {
#ifdef FBCON_HAS_MFB
        case 1:
                palette[0] |= (1 << 11);

                fbcon_cmap.mfb[regno] = pal;
                break;
#endif
#ifdef FBCON_HAS_CFB2
        case 2:
                palette[0] |= (1 << 12);

                fbcon_cmap.cfb2[regno] = pal;
                break;
#endif
#ifdef FBCON_HAS_CFB4
        case 4:
                palette[0] |= (2 << 12);

                fbcon_cmap.cfb4[regno] = pal;
                break;
#endif
#ifdef FBCON_HAS_CFB8
        case 8:
                palette[0] |= (3 << 12);

                fbcon_cmap.cfb8[regno] = pal;
                break;
#endif
#ifdef FBCON_HAS_CFB16
        case 16:
                palette[0] |= (4 << 12);

                fbcon_cmap.cfb16[regno] = pal;
                break;
#endif
        default:
                printk(KERN_ALERT "GLCD: Unsupported Video BPP %d!\n",current_par.bpp);
                break;
        }

	return 0;
}

/* Palette Initialization */
static void tifb_init_palette()
{
	unsigned short *palette = (unsigned short *)current_par.v_palette_base;
	unsigned short i, size;

	// Palette Size
	size = (current_par.palette_size/sizeof(*palette));

	// Clear the Palette
	memset(palette, 0, current_par.palette_size);

	// Initialization of Palette for Default values
	for (i = 0; i < size; i++){
		*(unsigned short *)(palette + i) = i;
	}

	// Setup the BPP
	switch ( current_par.bpp ) 
	{
#ifdef FBCON_HAS_MFB
	case 1:
	        palette[0] |= (1 << 11);
	        break;
#endif
#ifdef FBCON_HAS_CFB2
	case 2:
	        palette[0] |= (1 << 12);
	        break;
#endif
#ifdef FBCON_HAS_CFB4
	case 4:
	        palette[0] |= (2 << 12);
	        break;
#endif
#ifdef FBCON_HAS_CFB8
	case 8:
	        palette[0] |= (3 << 12);
	        break;
#endif
#ifdef FBCON_HAS_CFB16
	case 16:
	        palette[0] |= (4 << 12);
	        break;
#endif
	default:
	        printk(KERN_ALERT "GLCD: Unsupported Video BPP %d!\n",current_par.bpp);
	        break;
	}

#ifdef FB_DEBUG
	printk(KERN_INFO "GLCD: Palette initialization done successfully\n");
#endif

	// Clear the Frame Buffer Memory
	memset(current_par.v_screen_base, 0x00, current_par.screen_size);
}

/* Map Video Memory */
static int tifb_map_video_memory(void)
{
	struct page *page;

	/* Set Reserved Flag for FB Memory to allow it to be Remapped into
	 * User Space by the fbmem Driver */
	for ( page = virt_to_page(current_par.v_screen_base);
		page < virt_to_page(current_par.v_screen_base +
			((current_par.screen_size + 0xffe) & 0xfffff000));
		page++) {
				mem_map_reserve(page);
	}

	memset(current_par.v_screen_base, 0, current_par.screen_size);

#ifdef FB_DEBUG
	printk(KERN_INFO "GLCD: Video Memory mapping successful\n");
#endif

	return 1;
}


/* Initialize the Frame Buffer */
static __init int tifb_init(void)
{
        char *buf;
        
        if ( !tifb_enable )
                return -ENXIO;

        printk(KERN_INFO "GLCD: Initializing driver\n");

        /* Update the lcd_cfg struct. */
        if (STATUS_SUCCESS != lcd_hal_init(&lcd_cfg)) {
        	return -EFAULT;
        }
        
        buf_size = ( lcd_cfg.p_disp_panel->height * lcd_cfg.p_disp_panel->width * lcd_cfg.bpp ) / 8;

	      // Allocate the Memory for Frame Buffer
        if ( ( buf = kmalloc( buf_size, GFP_DMA ) ) == NULL )
        {
	         printk(KERN_ALERT "GLCD: Unable to allocate Memory for Frame Buffer");
	         return -EINVAL;
	      }

	      current_par.screen_size = buf_size;
	      current_par.p_screen_base  = (unsigned char *) virt_to_phys(buf);
	      current_par.v_screen_base  = buf;
	      tifb_memory = (u_long)buf;

	      // Allocate the Memory for Palette, maximum 256 entries of 16-bit
	      if ( ( buf = kmalloc( PALETTE_SIZE * 2, GFP_DMA ) ) == NULL )
        {       
	         printk(KERN_ALERT "GLCD: Unable to allocate Memory for palette");
	         return -EINVAL;
	      }

	      current_par.palette_size = PALETTE_SIZE;
	      current_par.p_palette_base = (unsigned char *) virt_to_phys(buf);
	      current_par.v_palette_base = buf;
       
#ifdef FB_DEBUG
	printk(KERN_INFO "GLCD: Frame Buffer Memory allocation successful\n");
        printk(KERN_INFO "GLCD: Virtural Screen Base allocated by kmalloc() is 0x%lx\n",(unsigned long)buf);
#endif

        // Initialize fbinfo
        strcpy(tifb_info.modename, desc);
        strcpy(tifb_info.fontname, "Acorn8x8");
        tifb_info.node       = -1;
        tifb_info.flags      = FBINFO_FLAG_DEFAULT;
        tifb_info.fbops      = &tifb_ops;
        tifb_info.disp       = &disp;
        tifb_info.switch_con = &tifbcon_switch;
        tifb_info.blank      = &tifbcon_blank;
        tifb_info.changevar  = NULL;

        // Initialize current_par
        current_par.bpp = lcd_cfg.bpp;
	      current_par.currcon	= -1;
        
        
        tifb_default.xres = lcd_cfg.p_disp_panel->width;
        tifb_default.xres_virtual = lcd_cfg.p_disp_panel->width;
        
        tifb_default.yres = lcd_cfg.p_disp_panel->height;
        tifb_default.yres_virtual = lcd_cfg.p_disp_panel->height;

        tifb_default.grayscale = lcd_cfg.p_disp_panel->panel_shade == MONOCROME ? 1 : 0;
        tifb_default.bits_per_pixel = lcd_cfg.bpp;
        
        tifb_default.hsync_len = lcd_cfg.hsw;
        tifb_default.vsync_len = lcd_cfg.vsw;
        
	      // Initialize the Palette
#ifdef FB_DEBUG
	      printk(KERN_INFO "GLCD: Initializing the Palette...\n");
#endif
	      tifb_init_palette();

     	// Map Video Memory
#ifdef FB_DEBUG
	      printk(KERN_INFO "GLCD: Mapping the Video Memory...\n");
#endif
	      tifb_map_video_memory();

	      /* Update the data buffer with any splash/welcome frame TODO-Later*/
	    
	      /* Set the palette: the following call should block till the respective
	       * interrupt and status bits are updated.
	       */
	      lcd_hal_blit(LOAD_PALETTE, current_par.p_palette_base);
    
	      /* Flush the buffer to the screen. */
	      lcd_hal_blit(LOAD_DATA, 
	      			( char *)current_par.p_screen_base );
       
        // initialize var_screeninfo
        tifb_set_var(&tifb_default, -1, &tifb_info);

        // Register the Frame Buffer
        if ( register_framebuffer(&tifb_info) < 0 ){
           printk(KERN_ALERT "GLCD: Frame Buffer Registration Failed!\n");
           kfree( current_par.v_palette_base );
           kfree( current_par.v_screen_base );
           return -EINVAL;
        }

	printk(KERN_INFO "GLCD: Registered the Frame Buffer successfully\n");

#ifdef FB_DEBUG

#ifdef CONFIG_GLCD_SHARP_COLOR
{
  /* Load a 16-bin color image */
  memcpy( current_par.v_screen_base, image, sizeof( image ) );

  printk(KERN_INFO "GLCD: Displaying a color image...\n");
}
#elif defined( CONFIG_GLCD_OPTREX_MONO )
{
	/* To print the default Palette Values and Update the FrameBuffer */
	unsigned short i, j, swidth, fb_size;
	unsigned short *pal_ptr, pal_size;
	unsigned char *data_ptr, pix_data1, pix_data2;

	pal_ptr = (unsigned short *)current_par.v_palette_base;
	pal_size = (current_par.palette_size/sizeof(*pal_ptr));
	fb_size = ((NUM_XPIXELS/2) * (NUM_YPIXELS));
	data_ptr = current_par.v_screen_base;
	swidth = 16;		// Strip Width (# of pixels in multiples of 16)
	pix_data1 = 0x00;
	i = 0;
	j = 0;

	/* Generation of Chess Board Pattern */
	data_ptr = current_par.v_screen_base;
	pix_data1 = 0x00;
	pix_data2 = 0x00;

	printk(KERN_INFO "GLCD: Generating Chess Board Pattern...\n");
	for(i = 0; i < (NUM_YPIXELS); i++){
		if((i % swidth) == 0)
			pix_data1 ^= 0xff;
		pix_data2 = ~pix_data1;
		for(j = 0; j < (NUM_XPIXELS/2); j++){
			if ((j % (swidth/2)) == 0)
            {
              if( pix_data2 )
                 pix_data2 = 0;
              else
              {
                 pix_data2 = (( j % 15 ) + 1) & 0xf;   
                 pix_data2 |= pix_data2 << 4;
              }
            }   
			*data_ptr = pix_data2;
			data_ptr++;
		}
	}
}
#endif
#endif

  return 0;
}

static void set_color_bitfields(struct fb_var_screeninfo *var)
{
    switch (var->bits_per_pixel) 
    {
	case 1:
	case 8:
	    var->red.offset = 0;
	    var->red.length = 8;
	    var->green.offset = 0;
	    var->green.length = 8;
	    var->blue.offset = 0;
	    var->blue.length = 8;
	    var->transp.offset = 0;
	    var->transp.length = 0;
	    break;
	case 4:
	    var->red.offset = 0;
	    var->red.length = 4;
	    var->green.offset = 0;
	    var->green.length = 4;
	    var->blue.offset = 0;
	    var->blue.length = 4;
	    var->transp.offset = 0;
	    var->transp.length = 0;
	    break;
	case 16:	/* RGB 565 */
	    var->red.offset = 0;
	    var->red.length = 5;
	    var->green.offset = 5;
	    var->green.length = 6;
	    var->blue.offset = 11;
	    var->blue.length = 5;
	    var->transp.offset = 0;
	    var->transp.length = 0;
	    break;
    }
    var->red.msb_right = 0;
    var->green.msb_right = 0;
    var->blue.msb_right = 0;
    var->transp.msb_right = 0;
}

/* Frame Buffer Exit Sequence */
static void __exit tifb_cleanup(void)
{
	// Un-register the Frame Buffer
	printk(KERN_INFO "GLCD: Frame Buffer Un-registering...\n");

	if(unregister_framebuffer(&tifb_info) < 0){
		printk(KERN_ALERT "GLCD: Error un-registering the Frame Buffer!\n");
	}
  
	// Free the Pages allocated by the Kernel
	printk(KERN_INFO "GLCD: Freeing the Memory allocated for Frame Buffer...\n");
  kfree( current_par.v_palette_base );
  kfree( current_par.v_screen_base );

	printk(KERN_INFO "GLCD: Frame Buffer clean-up successful\n");
}

module_init(tifb_init);
module_exit(tifb_cleanup);

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
