/**************************************************************************
 * FILE PURPOSE	:  TI Quality of Service	
 **************************************************************************
 * FILE NAME	:   ti_qos.h
 *
 * DESCRIPTION	:
 * 	Stuctures and definations used by the TI QOS Modules.
 *
 *	(C) Copyright 2004-2005, Texas Instruments, Inc.
 *************************************************************************/

#ifndef __TI_QOS__H__
#define __TI_QOS__H__

#include <linux/ip.h>

/******************************* DEFINITIONS *****************************/

/**************************************************************************
 * STRUCTURE -  TI_NSP_QOS_CLASS
 **************************************************************************
 *  This is the Class of Service that is maintained in the System. This
 *  is used only for the internal aggregation.
 **************************************************************************/
typedef enum TI_NSP_QOS_CLASS
{
    TI_NSP_QOS_COS1  = 0x0,     /* Highest Priority --> EF Traffic */
    TI_NSP_QOS_COS2  = 0x1,
    TI_NSP_QOS_COS3  = 0x2,
    TI_NSP_QOS_COS4  = 0x3,
    TI_NSP_QOS_COS5  = 0x4,
    TI_NSP_QOS_COS6  = 0x5,     /* Lowest Priority --> BE Traffic */
}TI_NSP_QOS_CLASS;

/* This is the MAX. Number of Class of Service Profiles that are present in
 * the QoS Framework. */
#define TI_NSP_SERVICE_NUMBER       0x6

/* This is the max. number of priorities that are supported in a VLAN packet. */
#define MAX_VLAN_PRIORITY           8

/* These are the definitions of the well known DSCP as per the RFC. */
#define DSCP_MGMT                   0x3F
#define DSCP_EF                     0x2E
#define DSCP_AF41                   0x22
#define DSCP_AF42                   0x24
#define DSCP_AF43                   0x26
#define DSCP_AF31                   0x1A
#define DSCP_AF32                   0x1C
#define DSCP_AF33                   0x1E
#define DSCP_AF21                   0x12
#define DSCP_AF22                   0x14
#define DSCP_AF23                   0x16
#define DSCP_AF11                   0xA
#define DSCP_AF12                   0xC
#define DSCP_AF13                   0xE
#define DSCP_BE                     0x0

/* The is the max. number of DSCP values that can be encoded in an IP packet. */
#define MAX_DSCP                    256

/* This definition controls the max. number of entries in the QoS database for 
 * prioritization of packet for local applications. */
#define MAX_NUMBER_LOCAL_PORTS      10

/****************************** FLAG Definitions *****************************/

/* This is the information that is carried in the packets. This is marked by
 * the various NSP QoS framework layers. */

/* This indicates if the packet has been classified and has passed through the
 * PIL Layer. All packets should be "Classified" before they hit the NQM. */
#define TI_QOS_PACKET_CLASSIFIED    0x1

/* This indicates that a packet has been translated by the Translation Unit.
 * All packets should be translated before they hit the "Shaper" */
#define TI_QOS_PACKET_TRANSLATED    0x2

/* This flag is used only for the Payload database design. Its set for fragmented
 * packets received. The query to the connection tracking database shall not 
 * suceed now because of fragmentation.  */
#define TI_QOS_PACKET_FRAGMENTED    0x4

/* This flag indicates that the Connection Tracking database was not able to 
 * determine the priority of the packet. This is generally the case for the first
 * packet of a connection. In such cases the packets are marked as "Best-Effort"
 * till they get marked accordingly in the "MANGLE" table. In such cases the
 * priority inside the packet is NOT correct. */
#define TI_QOS_PACKET_LIE           0x8

/**************************************************************************
 * STRUCTURE -  TI_QOS_LOCAL_PORT
 **************************************************************************
 *  This structure defines that the any incoming packet for "port" is 
 *  treated as a CoS1 packet. This structure and related API should be 
 *  used only for HIGH Priority TRUSTED applications like Voice.
 **************************************************************************/
typedef struct TI_QOS_LOCAL_PORT
{
    /* Links to the high priority ports */ 
    struct list_head        links;

    /* The port on which the "local" application will exchange data */
    unsigned int            port;
}TI_QOS_LOCAL_PORT;

/**************************************************************************
 * STRUCTURE -  TI_QOS_LOCAL_DATABASE_MCB
 **************************************************************************
 *  This structure is the MCB that contains all the information about the
 *  HIGH priority "port" database.
 **************************************************************************/
typedef struct TI_QOS_LOCAL_DATABASE_MCB
{
    /* List of all pre-allocated high priority ports. */
    struct list_head        available_port_list;

    /* List of active high priority ports. */
    struct list_head        active_port_list;
}TI_QOS_LOCAL_DATABASE_MCB;

/**************************************************************************
 * STRUCTURE -  TI_QOS_TCA_MODE
 **************************************************************************
 *  This specifies the "mode" of the TCA associated with the network device.
 *  The mode defines the type of network attached behind the interface. 
 **************************************************************************/
typedef enum TI_QOS_TCA_MODE
{
    TI_NSP_QOS_NONE             = 0x0,  /* Special Case:                    */
	TI_NSP_QOS_LAYER2_TRUSTED 	= 0x1,  /* Layer 2 --> Ingress Mappings     */
	TI_NSP_QOS_LAYER3_TRUSTED   = 0x2,  /* Layer 3 --> Ingress Mappings     */
    TI_NSP_QOS_STATIC           = 0x4,  /* Use the Source Interface rule    */
    TI_NSP_QOS_EGRESS_LAYER3    = 0x8,  /* Layer 3 --> Egress Mappings      */
    TI_NSP_QOS_EGRESS_LAYER2    = 0x10, /* Layer 2 --> Egress Mappings      */
}TI_QOS_TCA_MODE;

/**************************************************************************
 * STRUCTURE -  TI_QOS_PACKET_INFO
 **************************************************************************
 *  This structure contains the information that is carried in a per packet
 *  basis. This is hooked to the network buffer structure.
 **************************************************************************/
typedef struct TI_QOS_PACKET_INFO
{
    /* Generic Flag information used to carry a bitmask of information. */
    unsigned int                        ti_qos_flags;

    /* This is the internal aggregated Class of Service. */
    TI_NSP_QOS_CLASS                    ti_service_class;

    /* This is the TUPLE hash that is carried in the packet. If the 
     * connection tracking database is looked then this is a cached
     * copy that is carried forth so that the connection tracking
     * code need NOT lookup the database again. */
    struct ip_conntrack_tuple_hash*     ti_tuple_hash;
    struct ip_conntrack_protocol*       ti_connection_proto;
}TI_QOS_PACKET_INFO;

/**************************************************************************
 * STRUCTURE -  TI_QOS_TCA
 **************************************************************************
 *  This structure contains the TCA associated with each network interface. 
 *  When a packet is received on the network interface it uses the TCA to
 *  translate it to NSP Service Levels. Similary when a packet is 
 *  transmitted on the this network interface a translation is done from the
 *  NSP service levels to the TCA defined.
 **************************************************************************/
typedef struct TI_QOS_TCA
{
    /* The "mode" of the network attached to this interface. */
    TI_QOS_TCA_MODE         tca_mode;

    /* The pointer to the device on which this TCA has been attached. */
    struct net_device*      dev;    

    /* This is the Class of Service to which all packets will be mapped 
     * if the "Source Interface" Rule is configured. If the interface is
     * in the DYNAMIC mode then this is NOT used. */
    TI_NSP_QOS_CLASS        src_interface_service_class;

    /***********************************************************************
     ************************** LAYER 2 SECTION ****************************
     ***********************************************************************/

    /****************************** Ingress *******************************/
    
    /* The "Ingress" Mapping at Layer 2 maps a VLAN User priority to a Class
     * of Service */
    TI_NSP_QOS_CLASS       ingress_layer2_mapping[MAX_VLAN_PRIORITY];

    /*************************** Egress ***********************************/

    /* The "Egress" Mappings for Layer 2 maps the Class of Service to a 
     * known VLAN User Priority. */
    int                     egress_layer2_mapping[TI_NSP_SERVICE_NUMBER];

    /* This defines the service class to which "unclassified" packets at
     * Layer 2 are marked to. */
    TI_NSP_QOS_CLASS        default_egress_service_class;
    
    /***********************************************************************
     ************************** LAYER 3 SECTION ****************************
     ***********************************************************************/

    /****************************** Ingress *******************************/

    /* The "Ingress" Mapping at Layer 3 maps a DSCP to a Class of Service */
    TI_NSP_QOS_CLASS        ingress_layer3_mapping[MAX_DSCP];

    /* This is the default classification of non IP packet. PPP Control/ARP 
     * Frames are examples that could not be classified and fall into this
     * service level. */
    TI_NSP_QOS_CLASS        default_non_ip_ingress_service_class;

    /*************************** Egress ***********************************/

    /* The "Egress" Mapping maps a Class of Service Level to a DSCP. */
    unsigned char           egress_layer3_mapping[TI_NSP_SERVICE_NUMBER];

    /* This defines the service class to which "unclassified" NON IP packets 
     * are mapped to. */
    TI_NSP_QOS_CLASS        default_non_ip_egress_service_class;
    
    /* This defines the service class to which "unclassified" IP packets are 
     * mapped to. */
    TI_NSP_QOS_CLASS        default_ip_egress_service_class;
}TI_QOS_TCA;

/**************************************************************************
 ************************** Network Queue Manager *************************
 **************************************************************************/

/* These define the CONGESTION Levels that are present in the System. */
#define TI_QOS_NQM_NOT_CONGESTED    0x0
#define TI_QOS_NQM_DROP_COS1        0x1
#define TI_QOS_NQM_DROP_COS2        0x2
#define TI_QOS_NQM_DROP_COS3        0x3
#define TI_QOS_NQM_DROP_COS4        0x4
#define TI_QOS_NQM_DROP_COS5        0x5
#define TI_QOS_NQM_DROP_COS6        0x6

/**************************************************************************
 * STRUCTURE -  TI_QOS_NQM_MCB
 **************************************************************************
 *  This structure is the master control block that contains all the
 *  information about the Network Queue Manager. It contains information 
 *  about the type of algorithm that is being used and also necessary
 *  statistics.
 **************************************************************************/
typedef struct TI_QOS_NQM_MCB
{
    /* Name of the scheduler that is attached. */
    char*               name;

    /* The enqueue function pointer is called by the drivers to place a packet 
     * into a queue. */
    int                (*enqueue)(struct sk_buff *skb);

    /* This function is called to dequeue a packet from the NQM. The function
     * can return NULL indicating that there are no more packets available. */
    struct sk_buff*     (*dequeue)(void);

    /* Congestion Check API that needs to be provided. The function returns a
     * simple 1/0 value indicating if the packet should be dropped or
     * accepted. */
    int                 (*congestion_check)(struct sk_buff *skb, TI_NSP_QOS_CLASS service_class);

    /* Pointer to the private NQM algorithm structure. */
    void*               ptr_nqm_private_info;
}TI_QOS_NQM_MCB;

/**************************************************************************
 * STRUCTURE -  TI_QOS_NQM_RR
 **************************************************************************
 *  The structure contains information related to the Round-Robin NQM 
 *  algorithm.
 **************************************************************************/
typedef struct TI_QOS_NQM_RR
{
    /* The internal NSP Queues:-
     *  Indexed by the NSP Service Level specifies the queue that is to be used. */   
    struct sk_buff_head	        nsp_pkt_queue[TI_NSP_SERVICE_NUMBER];

    /* Threshold of the queue; if the queue length exceeds threshold we are
     * nearing congestion. This like above is indexed by the NSP Service
     * level. */
    int                         nsp_threshold[TI_NSP_SERVICE_NUMBER];

    /* This defines the congestion in the System. */
    int                         nsp_congestion_level;

    /* This bit mask defines the state of each of the queues. A 1 indicates
     * that the queue is FULL and 0 indicates that the queue is empty. */
    int                         nsp_queue_state;

    /* Statistics. */
    int                         nsp_receive_stats[TI_NSP_SERVICE_NUMBER];
    int                         nsp_dropped_packets[TI_NSP_SERVICE_NUMBER];
}TI_QOS_NQM_RR;

/**************************************************************************
 **************************** Translation Unit ****************************
 **************************************************************************/

/**************************************************************************
 * STRUCTURE -  TI_QOS_TU_MCB
 **************************************************************************
 *  This structure contains the information related to the Translation 
 *  Unit.
 **************************************************************************/
typedef struct TI_QOS_TU_MCB
{
    /* This counter specifies the number of IP packets which had their TOS
     * Byte changed. */ 
    int                     tos_byte_remarked;

    /* This specifies the number of DSCP packets that were not modified at all. */
    int                     pkts_unchanged;

    /* This specifies the number of non IP Packets that passed the TU. */
    int                     non_ip_pkts;

    /* This specifies the number of unclassified IP Packets that were passed
     * to the TU. */
    int                     unclassified_ip_pkts;

    /* This specifies the number of unclassified non-IP Packets that were passed
     * to the TU. */
    int                     unclassified_non_ip_pkts;

    /* This specifies the number of packets that were not classified and were 
     * passed to the Layer 2 Translation unit handler. */
    int                     unclassified_layer2_pkts;
}TI_QOS_TU_MCB;

/**************************************************************************
 * STRUCTURE -  TI_QOS_MCB
 **************************************************************************
 *  This structure contains all the structures that combine to form the
 *  NSP QoS Framework. There is only instance of this in the NSP System.
 **************************************************************************/
typedef struct TI_QOS_MCB
{
    /* Handle to the Network Queue Manager. */
    TI_QOS_NQM_MCB*                 ptr_nqm_handle;

    /* Handle to the Translation Unit. */
    TI_QOS_TU_MCB*                  ptr_tu_handle;

    /* Handle to the Local HIGH Priority Port Database. */
    TI_QOS_LOCAL_DATABASE_MCB*      ptr_local_database;

    /* Statistics.... */
    int                             classification_error;
    int                             unclassified_pkts;
    int                             fragmented_pkts;
}TI_QOS_MCB;

/******************************* EXTERN FUNCTIONS *****************************/

/********************************* TU Exports ********************************/
extern int ti_qos_layer2_translation_unit (struct sk_buff *skb);
extern void ti_qos_translation_unit(struct sk_buff *skb, struct net_device* dev);

/********************************* PIL Exports ********************************/
extern int ti_pil_classify_ethernet_frame (struct sk_buff* skb, int protocol_offset);
extern int ti_pil_classify_ip_packet (struct sk_buff* skb, struct iphdr* ptr_iphdr);
extern int ti_pil_classify_wlan_frame (struct sk_buff* skb, struct net_device* dev);
extern int ti_pil_classify_wlan_eapol_frame (struct sk_buff* skb);

extern int ti_qos_register_local_port (unsigned int port);
extern int ti_qos_deregister_local_port (int port);

/********************************* NQM Round-Robin Algorithm ******************/
extern int ti_nqm_rr_enqueue_packet (struct sk_buff* skb);
extern struct sk_buff* ti_nqm_rr_dequeue_packet (void);
extern int ti_nqm_rr_congestion_check (struct sk_buff* skb, TI_NSP_QOS_CLASS service_class);

/********************************* Init/Deinit Functions **********************/
extern int ti_nsp_qos_initialize (void);
extern int ti_tca_device_init(struct net_device *dev);
extern int ti_tca_device_deinit(struct net_device *dev);

/********************************* GLOBAL Variable ****************************/

/* This is the handle to the QoS MASTER CONTROL Block that has all the
 * information. */
extern TI_QOS_MCB* ptr_ti_qos_mcb;

/* This is the global variable that controls the QoS framework. */
extern int gIsQoSEnabled;

#endif /* __TI_QOS__H__ */
