/***************************************************************************\
*****************************************************************************
**                                                                         **
** source file:    memory.c                                                **
** last revision:  18/01/2004                                              **
**                                                                         **
*****************************************************************************
*****************************************************************************
**                                                                         **
** This program is free software; you can redistribute it and/or modify it **
** under the terms of the GNU General Public License as published by the   **
** Free Software Foundation; either version 2, or (at your option) any     **
** later version.                                                          **
**                                                                         **
** This program is distributed in the hope that it will be useful,         **
** but WITHOUT ANY WARRANTY; without even the implied warranty of          **
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           **
** GNU General Public License for more details.                            **
**                                                                         **
** You should have received a copy of the GNU General Public License       **
** along with this program; if not, write to the Free Software             **
** Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.                     **
**                                                                         **
*****************************************************************************
\***************************************************************************/

//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include "error.h"
#include "memory.h"

//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------
// This Function Creates A Memory Manager

MEM_MANAGER *mem_create(void)
{
  MEM_MANAGER *mem;
  mem = (MEM_MANAGER *) malloc(sizeof(MEM_MANAGER));

  if (mem == NULL)
    {
      error("mem_init", "unable to allocate memory", 0);
      return NULL;
    }

  mem->head = NULL;
  mem->tail = NULL;
  return mem;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------
// This Re-Creates A Memory Manager

MEM_MANAGER *mem_recreate(MEM_MANAGER *mem)
{
  if (mem == NULL)
    return NULL;

  mem_destroy(mem);
  return mem_create();
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------
// This Allocates Some Memory

int mem_alloc(MEM_MANAGER *mem, void **datas, const size_t size)
{
  MEM_BRICK *tmp = NULL;

  if (mem == NULL)
    return -1;

  // Chained List Behavior
  if (mem->head == NULL)
    {
      mem->head = malloc(sizeof(MEM_BRICK));

      if (mem->head == NULL)
        {
          error("mem_alloc", "unable to allocate memory", 0);
          return -1;
        }

      mem->tail = mem->head;
    }
  else
    {
      tmp = mem->tail;
      mem->tail = malloc(sizeof(MEM_BRICK));

      if (mem->tail == NULL)
        {
          error("mem_alloc", "unable to allocate memory", 0);
          return -1;
        }

      tmp->next_brick = mem->tail;
    }

  // Datas Storage Allocation
  *datas = malloc(size);

  if (*datas == NULL)
    {
      error("mem_alloc", "unable to allocate memory", 0);
      return -1;
    }

  mem->tail->prev_brick = tmp;
  mem->tail->next_brick = NULL;
  mem->tail->datas = *datas;
  mem->tail->size = size;
  return 0;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------
// This Reallocates Some Memory

int mem_realloc(MEM_MANAGER *mem, void **datas, const size_t size)
{
  MEM_BRICK *tmp;

  if (mem == NULL)
    return -1;

  if (mem->head == NULL)
    return -1;

  for (tmp=mem->head; tmp!=NULL; tmp=tmp->next_brick)
    {
      if (tmp->datas == *datas)
        {
          *datas = realloc(*datas, size);

          if (*datas == NULL)
            {
              error("mem_realloc", "unable to reallocate memory", 0);
              return -1;
            }

          tmp->datas = *datas;
          tmp->size = size;
          return 0;
        }
    }

  return -1;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------
// This Returns The Amount Of Allocated Memory

size_t mem_sizeof(MEM_MANAGER *mem, const void *datas)
{
  MEM_BRICK *tmp;

  if (mem == NULL)
    return -1;

  if (mem->head == NULL)
    return -1;

  for (tmp=mem->head; tmp!=NULL; tmp=tmp->next_brick)
    if (tmp->datas == datas)
      return tmp->size;

  return 0;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------
// This Frees The Allocated Memory Of An Element

int mem_free(MEM_MANAGER *mem, void **datas)
{
  MEM_BRICK *tmp;

  if (mem == NULL)
    return -1;

  if (mem->head == NULL)
    return -1;

  for (tmp=mem->head; (tmp != NULL); tmp=tmp->next_brick)
    {
      if (tmp->datas == *datas)
        {
          if (tmp->next_brick != NULL)
            tmp->next_brick->prev_brick = tmp->prev_brick;

          if (tmp->prev_brick != NULL)
            tmp->prev_brick->next_brick = tmp->next_brick;

          if (tmp == mem->head)
            mem->head = mem->head->next_brick;

          if (tmp == mem->tail)
            mem->tail = mem->tail->prev_brick;

          free(*datas);
          free(tmp);
          return 0;
        }
    }

  return -1;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------
// This Frees All The Allocations Of A Memory Manager

void mem_free_all(MEM_MANAGER *mem)
{
  MEM_BRICK *tmp;

  if (mem == NULL)
    return;

  if (mem->head == NULL)
    return;

  for (tmp=mem->tail; (tmp != NULL); tmp=tmp->prev_brick)
    {
      free(tmp->datas);
      free(tmp->next_brick);
    }

  free(mem->head);
  mem->head = NULL;
  mem->tail = NULL;
  return;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------
// This Destroys A Memory Manager

void mem_destroy(MEM_MANAGER *mem)
{
  if (mem == NULL)
    return;

  mem_free_all(mem);
  free(mem);
  return;
}


//---------------------------------------------------------------------------------------
//|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//---------------------------------------------------------------------------------------
